#
# Copyright 2017 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import performance as perf
from . import utils
from functools import wraps
from matplotlib.ticker import ScalarFormatter
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import pandas as pd
from scipy import stats
import statsmodels.api as sm

import seaborn as sns
import matplotlib
# matplotlib.use('TkAgg')


DECIMAL_TO_BPS = 10000


def customize(func):
    """
    Decorator to set plotting context and axes style during function call.
    """
    @wraps(func)
    def call_w_context(*args, **kwargs):
        set_context = kwargs.pop('set_context', True)
        if set_context:
            color_palette = sns.color_palette('colorblind')
            with plotting_context(), axes_style(), color_palette:
                sns.despine(left=True)
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return call_w_context


def plotting_context(context='notebook', font_scale=1.5, rc=None):
    """
    Create alphalens default plotting style context.

    Under the hood, calls and returns seaborn.plotting_context() with
    some custom settings. Usually you would use in a with-context.

    Parameters
    ----------
    context : str, optional
        Name of seaborn context.
    font_scale : float, optional
        Scale font by factor font_scale.
    rc : dict, optional
        Config flags.
        By default, {'lines.linewidth': 1.5}
        is being used and will be added to any
        rc passed in, unless explicitly overriden.

    Returns
    -------
    seaborn plotting context

    Example
    -------
    with alphalens.plotting.plotting_context(font_scale=2):
        alphalens.create_full_tear_sheet(..., set_context=False)

    See also
    --------
    For more information, see seaborn.plotting_context().
    """
    if rc is None:
        rc = {}

    rc_default = {'lines.linewidth': 1.5}

    # Add defaults if they do not exist
    for name, val in rc_default.items():
        rc.setdefault(name, val)

    return sns.plotting_context(context=context, font_scale=font_scale, rc=rc)

def axes_style(style='darkgrid', rc=None):
    """Create alphalens default axes style context.

    Under the hood, calls and returns seaborn.axes_style() with
    some custom settings. Usually you would use in a with-context.

    Parameters
    ----------
    style : str, optional
        Name of seaborn style.
    rc : dict, optional
        Config flags.

    Returns
    -------
    seaborn plotting context

    Example
    -------
    with alphalens.plotting.axes_style(style='whitegrid'):
        alphalens.create_full_tear_sheet(..., set_context=False)

    See also
    --------
    For more information, see seaborn.plotting_context().

    """
    if rc is None:
        rc = {}

    rc_default = {}

    # Add defaults if they do not exist
    for name, val in rc_default.items():
        rc.setdefault(name, val)

    return sns.axes_style(style=style, rc=rc)


def plot_returns_table(alpha_beta,
                       mean_ret_quantile,
                       mean_ret_spread_quantile):
    returns_table = pd.DataFrame()
    returns_table = returns_table.append(alpha_beta)
    returns_table.loc["Mean Period Wise Return Top Quantile (bps)"] = \
        mean_ret_quantile.iloc[-1] * DECIMAL_TO_BPS
    returns_table.loc["Mean Period Wise Return Bottom Quantile (bps)"] = \
        mean_ret_quantile.iloc[0] * DECIMAL_TO_BPS
    returns_table.loc["Mean Period Wise Spread (bps)"] = \
        mean_ret_spread_quantile.mean() * DECIMAL_TO_BPS

    print("Returns Analysis")
    utils.print_table(returns_table.apply(lambda x: x.round(3)))


def plot_turnover_table(autocorrelation_data, quantile_turnover):
    turnover_table = pd.DataFrame()
    for period in sorted(quantile_turnover.keys()):
        for quantile, p_data in quantile_turnover[period].iteritems():
            turnover_table.loc["Quantile {} Mean Turnover ".format(quantile),
                               "{}D".format(period)] = p_data.mean()
    auto_corr = pd.DataFrame()
    for period, p_data in autocorrelation_data.iteritems():
        auto_corr.loc["Mean Factor Rank Autocorrelation",
                      "{}D".format(period)] = p_data.mean()

    print("Turnover Analysis")
    utils.print_table(turnover_table.apply(lambda x: x.round(3)))
    utils.print_table(auto_corr.apply(lambda x: x.round(3)))


def plot_information_table(ic_data):
    ic_summary_table = pd.DataFrame()
    ic_summary_table["IC Mean"] = ic_data.mean()
    ic_summary_table["IC Std."] = ic_data.std()
    ic_summary_table["Risk-Adjusted IC"] = \
        ic_data.mean() / ic_data.std()
    t_stat, p_value = stats.ttest_1samp(ic_data, 0)
    ic_summary_table["t-stat(IC)"] = t_stat
    ic_summary_table["p-value(IC)"] = p_value
    ic_summary_table["IC Skew"] = stats.skew(ic_data)
    ic_summary_table["IC Kurtosis"] = stats.kurtosis(ic_data)

    print("Information Analysis")
    utils.print_table(ic_summary_table.apply(lambda x: x.round(3)).T)


def plot_quantile_statistics_table(factor_data):
    quantile_stats = factor_data.groupby('factor_quantile') \
        .agg(['min', 'max', 'mean', 'std', 'count'])['factor']
    quantile_stats['count %'] = quantile_stats['count'] \
        / quantile_stats['count'].sum() * 100.

    print("Quantiles Statistics")
    utils.print_table(quantile_stats)


def plot_ic_ts(ic, ax=None):
    """
    Plots Spearman Rank Information Coefficient and IC moving
    average for a given factor.

    Parameters
    ----------
    ic : pd.DataFrame
        DataFrame indexed by date, with IC for each forward return.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    ic = ic.copy()

    num_plots = len(ic.columns)
    if ax is None:
        f, ax = plt.subplots(num_plots, 1, figsize=(18, num_plots * 7))
        ax = np.asarray([ax]).flatten()

    ymin, ymax = (None, None)
    for a, (period_num, ic) in zip(ax, ic.iteritems()):
        ic.plot(alpha=0.7, ax=a, lw=0.7, color='steelblue')
        ic.rolling(window=22).mean().plot(
            ax=a,
            color='forestgreen',
            lw=2,
            alpha=0.8
        )

        a.set(ylabel='IC', xlabel="")
        a.set_title(
            "{} Period Forward Return Information Coefficient (IC)"
            .format(period_num))
        a.axhline(0.0, linestyle='-', color='black', lw=1, alpha=0.8)
        a.legend(['IC', '1 month moving avg'], loc='upper right')
        a.text(.05, .95, "Mean %.3f \n Std. %.3f" % (ic.mean(), ic.std()),
               fontsize=16,
               bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
               transform=a.transAxes,
               verticalalignment='top')

        curr_ymin, curr_ymax = a.get_ylim()
        ymin = curr_ymin if ymin is None else min(ymin, curr_ymin)
        ymax = curr_ymax if ymax is None else max(ymax, curr_ymax)

    for a in ax:
        a.set_ylim([ymin, ymax])

    return ax


def plot_ic_hist(ic, ax=None):
    """
    Plots Spearman Rank Information Coefficient histogram for a given factor.

    Parameters
    ----------
    ic : pd.DataFrame
        DataFrame indexed by date, with IC for each forward return.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    ic = ic.copy()

    num_plots = len(ic.columns)

    v_spaces = ((num_plots - 1) // 3) + 1

    if ax is None:
        f, ax = plt.subplots(v_spaces, 3, figsize=(18, v_spaces * 6))
        ax = ax.flatten()

    for a, (period_num, ic) in zip(ax, ic.iteritems()):
        sns.distplot(ic.replace(np.nan, 0.), norm_hist=True, ax=a)
        a.set(title="%s Period IC" % period_num, xlabel='IC')
        a.set_xlim([-1, 1])
        a.text(.05, .95, "Mean %.3f \n Std. %.3f" % (ic.mean(), ic.std()),
               fontsize=16,
               bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
               transform=a.transAxes,
               verticalalignment='top')
        a.axvline(ic.mean(), color='w', linestyle='dashed', linewidth=2)

    if num_plots < len(ax):
        ax[-1].set_visible(False)

    return ax


def plot_ic_qq(ic, theoretical_dist=stats.norm, ax=None):
    """
    Plots Spearman Rank Information Coefficient "Q-Q" plot relative to
    a theoretical distribution.

    Parameters
    ----------
    ic : pd.DataFrame
        DataFrame indexed by date, with IC for each forward return.
    theoretical_dist : scipy.stats._continuous_distns
        Continuous distribution generator. scipy.stats.norm and
        scipy.stats.t are popular options.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    ic = ic.copy()

    num_plots = len(ic.columns)

    v_spaces = ((num_plots - 1) // 3) + 1

    if ax is None:
        f, ax = plt.subplots(v_spaces, 3, figsize=(18, v_spaces * 6))
        ax = ax.flatten()

    if isinstance(theoretical_dist, stats.norm.__class__):
        dist_name = 'Normal'
    elif isinstance(theoretical_dist, stats.t.__class__):
        dist_name = 'T'
    else:
        dist_name = 'Theoretical'

    for a, (period_num, ic) in zip(ax, ic.iteritems()):
        sm.qqplot(ic.replace(np.nan, 0.).values, theoretical_dist, fit=True,
                  line='45', ax=a)
        a.set(title="{} Period IC {} Dist. Q-Q".format(
              period_num, dist_name),
              ylabel='Observed Quantile',
              xlabel='{} Distribution Quantile'.format(dist_name))

    return ax


def plot_quantile_returns_bar(mean_ret_by_q,
                              by_group=False,
                              ylim_percentiles=None,
                              ax=None):
    """
    Plots mean period wise returns for factor quantiles.

    Parameters
    ----------
    mean_ret_by_q : pd.DataFrame
        DataFrame with quantile, (group) and mean period wise return values.
    by_group : bool
        Disaggregated figures by group.
    ylim_percentiles : tuple of integers
        Percentiles of observed data to use as y limits for plot.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    mean_ret_by_q = mean_ret_by_q.copy()

    if ylim_percentiles is not None:
        ymin = (np.nanpercentile(mean_ret_by_q.values,
                                 ylim_percentiles[0]) * DECIMAL_TO_BPS)
        ymax = (np.nanpercentile(mean_ret_by_q.values,
                                 ylim_percentiles[1]) * DECIMAL_TO_BPS)
    else:
        ymin = None
        ymax = None

    if by_group:
        num_group = len(
            mean_ret_by_q.index.get_level_values('group').unique())

        if ax is None:
            v_spaces = ((num_group - 1) // 2) + 1
            f, ax = plt.subplots(v_spaces, 2, sharex=False,
                                 sharey=True, figsize=(18, 6 * v_spaces))
            ax = ax.flatten()

        for a, (sc, cor) in zip(ax, mean_ret_by_q.groupby(level='group')):
            (cor.xs(sc, level='group')
                .multiply(DECIMAL_TO_BPS)
                .plot(kind='bar', title=sc, ax=a))

            a.set(xlabel='', ylabel='Mean Return (bps)',
                  ylim=(ymin, ymax))

        if num_group < len(ax):
            ax[-1].set_visible(False)

        return ax

    else:
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(18, 6))

        (mean_ret_by_q.multiply(DECIMAL_TO_BPS)
            .plot(kind='bar',
                  title="Mean Period Wise Return By Factor Quantile", ax=ax))
        ax.set(xlabel='', ylabel='Mean Return (bps)',
               ylim=(ymin, ymax))

        return ax


def plot_quantile_returns_violin(return_by_q,
                                 ylim_percentiles=None,
                                 ax=None):
    """
    Plots a violin box plot of period wise returns for factor quantiles.

    Parameters
    ----------
    return_by_q : pd.DataFrame - MultiIndex
        DataFrame with date and quantile as rows MultiIndex,
        forward return windows as columns, returns as values.
    ylim_percentiles : tuple of integers
        Percentiles of observed data to use as y limits for plot.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    return_by_q = return_by_q.copy()

    if ylim_percentiles is not None:
        ymin = (np.nanpercentile(return_by_q.values,
                                 ylim_percentiles[0]) * DECIMAL_TO_BPS)
        ymax = (np.nanpercentile(return_by_q.values,
                                 ylim_percentiles[1]) * DECIMAL_TO_BPS)
    else:
        ymin = None
        ymax = None

    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    unstacked_dr = (return_by_q
                    .multiply(DECIMAL_TO_BPS))
    unstacked_dr.columns = unstacked_dr.columns.set_names('forward_periods')
    unstacked_dr = unstacked_dr.stack()
    unstacked_dr.name = 'return'
    unstacked_dr = unstacked_dr.reset_index()

    sns.violinplot(data=unstacked_dr,
                   x='factor_quantile',
                   hue='forward_periods',
                   y='return',
                   orient='v',
                   cut=0,
                   inner='quartile',
                   ax=ax)
    ax.set(xlabel='', ylabel='Return (bps)',
           title="Period Wise Return By Factor Quantile",
           ylim=(ymin, ymax))

    ax.axhline(0.0, linestyle='-', color='black', lw=0.7, alpha=0.6)

    return ax


def plot_mean_quantile_returns_spread_time_series(mean_returns_spread,
                                                  std_err=None,
                                                  bandwidth=1,
                                                  ax=None):
    """
    Plots mean period wise returns for factor quantiles.

    Parameters
    ----------
    mean_returns_spread : pd.Series
        Series with difference between quantile mean returns by period.
    std_err : pd.Series
        Series with standard error of difference between quantile
        mean returns each period.
    bandwidth : float
        Width of displayed error bands in standard deviations.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    if isinstance(mean_returns_spread, pd.DataFrame):
        if ax is None:
            ax = [None for a in mean_returns_spread.columns]

        ymin, ymax = (None, None)
        for (i, a), (name, fr_column) in zip(enumerate(ax),
                                             mean_returns_spread.iteritems()):
            stdn = None if std_err is None else std_err[name]
            a = plot_mean_quantile_returns_spread_time_series(fr_column,
                                                              std_err=stdn,
                                                              ax=a)
            ax[i] = a
            curr_ymin, curr_ymax = a.get_ylim()
            ymin = curr_ymin if ymin is None else min(ymin, curr_ymin)
            ymax = curr_ymax if ymax is None else max(ymax, curr_ymax)

        for a in ax:
            a.set_ylim([ymin, ymax])

        return ax

    if mean_returns_spread.isnull().all():
        return ax

    periods = mean_returns_spread.name
    title = ('Top Minus Bottom Quantile Mean Return ({} Period Forward Return)'
             .format(periods if periods is not None else ""))

    if ax is None:
        f, ax = plt.subplots(figsize=(18, 6))

    mean_returns_spread_bps = mean_returns_spread * DECIMAL_TO_BPS

    mean_returns_spread_bps.plot(alpha=0.4, ax=ax, lw=0.7, color='forestgreen')
    mean_returns_spread_bps.rolling(window=22).mean().plot(
        color='orangered',
        alpha=0.7,
        ax=ax
    )
    ax.legend(['mean returns spread', '1 month moving avg'], loc='upper right')

    if std_err is not None:
        std_err_bps = std_err * DECIMAL_TO_BPS
        upper = mean_returns_spread_bps.values + (std_err_bps * bandwidth)
        lower = mean_returns_spread_bps.values - (std_err_bps * bandwidth)
        ax.fill_between(mean_returns_spread.index,
                        lower,
                        upper,
                        alpha=0.3,
                        color='steelblue')

    ylim = np.nanpercentile(abs(mean_returns_spread_bps.values), 95)
    ax.set(ylabel='Difference In Quantile Mean Return (bps)',
           xlabel='',
           title=title,
           ylim=(-ylim, ylim))
    ax.axhline(0.0, linestyle='-', color='black', lw=1, alpha=0.8)

    return ax


def plot_ic_by_group(ic_group, ax=None):
    """
    Plots Spearman Rank Information Coefficient for a given factor over
    provided forward returns. Separates by group.

    Parameters
    ----------
    ic_group : pd.DataFrame
        group-wise mean period wise returns.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
    ic_group.plot(kind='bar', ax=ax)

    ax.set(title="Information Coefficient By Group", xlabel="")
    ax.set_xticklabels(ic_group.index, rotation=45)

    return ax


def plot_factor_rank_auto_correlation(factor_autocorrelation,
                                      period=1,
                                      ax=None):
    """
    Plots factor rank autocorrelation over time.
    See factor_rank_autocorrelation for more details.

    Parameters
    ----------
    factor_autocorrelation : pd.Series
        Rolling 1 period (defined by time_rule) autocorrelation
        of factor values.
    period: int, optional
        Period over which the autocorrelation is calculated
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    factor_autocorrelation.plot(title='{}D Period Factor Rank Autocorrelation'
                                .format(period), ax=ax)
    ax.set(ylabel='Autocorrelation Coefficient', xlabel='')
    ax.axhline(0.0, linestyle='-', color='black', lw=1)
    ax.text(.05, .95, "Mean %.3f" % factor_autocorrelation.mean(),
            fontsize=16,
            bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
            transform=ax.transAxes,
            verticalalignment='top')

    return ax


def plot_top_bottom_quantile_turnover(quantile_turnover, period=1, ax=None):
    """
    Plots period wise top and bottom quantile factor turnover.

    Parameters
    ----------
    quantile_turnover: pd.Dataframe
        Quantile turnover (each DataFrame column a quantile).
    period: int, optional
        Period over which to calculate the turnover.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    max_quantile = quantile_turnover.columns.max()
    min_quantile = quantile_turnover.columns.min()
    turnover = pd.DataFrame()
    turnover['top quantile turnover'] = quantile_turnover[max_quantile]
    turnover['bottom quantile turnover'] = quantile_turnover[min_quantile]
    turnover.plot(title='{}D Period Top and Bottom Quantile Turnover'
                  .format(period), ax=ax, alpha=0.6, lw=0.8)
    ax.set(ylabel='Proportion Of Names New To Quantile', xlabel="")

    return ax


def plot_monthly_ic_heatmap(mean_monthly_ic, ax=None):
    """
    Plots a heatmap of the information coefficient or returns by month.

    Parameters
    ----------
    mean_monthly_ic : pd.DataFrame
        The mean monthly IC for N periods forward.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    mean_monthly_ic = mean_monthly_ic.copy()

    num_plots = len(mean_monthly_ic.columns)

    v_spaces = ((num_plots - 1) // 3) + 1

    if ax is None:
        f, ax = plt.subplots(v_spaces, 3, figsize=(18, v_spaces * 6))
        ax = ax.flatten()

    new_index_year = []
    new_index_month = []
    for date in mean_monthly_ic.index:
        new_index_year.append(date.year)
        new_index_month.append(date.month)

    mean_monthly_ic.index = pd.MultiIndex.from_arrays(
        [new_index_year, new_index_month],
        names=["year", "month"])

    for a, (periods_num, ic) in zip(ax, mean_monthly_ic.iteritems()):

        sns.heatmap(
            ic.unstack(),
            annot=True,
            alpha=1.0,
            center=0.0,
            annot_kws={"size": 7},
            linewidths=0.01,
            linecolor='white',
            cmap=cm.coolwarm_r,
            cbar=False,
            ax=a)
        a.set(ylabel='', xlabel='')

        a.set_title("Monthly Mean {} Period IC".format(periods_num))

    if num_plots < len(ax):
        ax[-1].set_visible(False)

    return ax


def plot_cumulative_returns(factor_returns,
                            period,
                            freq=None,
                            title=None,
                            ax=None):
    """
    Plots the cumulative returns of the returns series passed in.

    Parameters
    ----------
    factor_returns : pd.Series
        Period wise returns of dollar neutral portfolio weighted by factor
        value.
    period : pandas.Timedelta or string
        Length of period for which the returns are computed (e.g. 1 day)
        if 'period' is a string it must follow pandas.Timedelta constructor
        format (e.g. '1 days', '1D', '30m', '3h', '1D1h', etc)
    freq : pandas DateOffset
        Used to specify a particular trading calendar e.g. BusinessDay or Day
        Usually this is inferred from utils.infer_trading_calendar, which is
        called by either get_clean_factor_and_forward_returns or
        compute_forward_returns
    title: string, optional
        Custom title
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    factor_returns = perf.cumulative_returns(factor_returns)

    factor_returns.plot(ax=ax, lw=3, color='forestgreen', alpha=0.6)
    ax.set(ylabel='Cumulative Returns',
           title=("Portfolio Cumulative Return ({} Fwd Period)".format(period)
                  if title is None else title),
           xlabel='')
    ax.axhline(1.0, linestyle='-', color='black', lw=1)

    return ax


def plot_cumulative_returns_by_quantile(quantile_returns,
                                        period,
                                        freq=None,
                                        ax=None):
    """
    Plots the cumulative returns of various factor quantiles.

    Parameters
    ----------
    quantile_returns : pd.DataFrame
        Returns by factor quantile
    period : pandas.Timedelta or string
        Length of period for which the returns are computed (e.g. 1 day)
        if 'period' is a string it must follow pandas.Timedelta constructor
        format (e.g. '1 days', '1D', '30m', '3h', '1D1h', etc)
    freq : pandas DateOffset
        Used to specify a particular trading calendar e.g. BusinessDay or Day
        Usually this is inferred from utils.infer_trading_calendar, which is
        called by either get_clean_factor_and_forward_returns or
        compute_forward_returns
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
    """

    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    ret_wide = quantile_returns.unstack('factor_quantile')

    cum_ret = ret_wide.apply(perf.cumulative_returns)

    cum_ret = cum_ret.loc[:, ::-1]  # we want negative quantiles as 'red'

    cum_ret.plot(lw=2, ax=ax, cmap=cm.coolwarm)
    ax.legend()
    ymin, ymax = cum_ret.min().min(), cum_ret.max().max()
    ax.set(ylabel='Log Cumulative Returns',
           title='''Cumulative Return by Quantile
                    ({} Period Forward Return)'''.format(period),
           xlabel='',
           yscale='symlog',
           yticks=np.linspace(ymin, ymax, 5),
           ylim=(ymin, ymax))

    ax.yaxis.set_major_formatter(ScalarFormatter())
    ax.axhline(1.0, linestyle='-', color='black', lw=1)

    return ax


def plot_quantile_average_cumulative_return(avg_cumulative_returns,
                                            by_quantile=False,
                                            std_bar=False,
                                            title=None,
                                            ax=None):
    """
    Plots sector-wise mean daily returns for factor quantiles
    across provided forward price movement columns.

    Parameters
    ----------
    avg_cumulative_returns: pd.Dataframe
        The format is the one returned by
        performance.average_cumulative_return_by_quantile
    by_quantile : boolean, optional
        Disaggregated figures by quantile (useful to clearly see std dev bars)
    std_bar : boolean, optional
        Plot standard deviation plot
    title: string, optional
        Custom title
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
    """

    avg_cumulative_returns = avg_cumulative_returns.multiply(DECIMAL_TO_BPS)
    quantiles = len(avg_cumulative_returns.index.levels[0].unique())
    palette = [cm.coolwarm(i) for i in np.linspace(0, 1, quantiles)]
    palette = palette[::-1]  # we want negative quantiles as 'red'

    if by_quantile:

        if ax is None:
            v_spaces = ((quantiles - 1) // 2) + 1
            f, ax = plt.subplots(v_spaces, 2, sharex=False,
                                 sharey=False, figsize=(18, 6 * v_spaces))
            ax = ax.flatten()

        for i, (quantile, q_ret) in enumerate(avg_cumulative_returns
                                              .groupby(level='factor_quantile')
                                              ):

            mean = q_ret.loc[(quantile, 'mean')]
            mean.name = 'Quantile ' + str(quantile)
            mean.plot(ax=ax[i], color=palette[i])
            ax[i].set_ylabel('Mean Return (bps)')

            if std_bar:
                std = q_ret.loc[(quantile, 'std')]
                ax[i].errorbar(std.index, mean, yerr=std,
                               fmt='none', ecolor=palette[i], label='none')

            ax[i].axvline(x=0, color='k', linestyle='--')
            ax[i].legend()
            i += 1

    else:

        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(18, 6))

        for i, (quantile, q_ret) in enumerate(avg_cumulative_returns
                                              .groupby(level='factor_quantile')
                                              ):

            mean = q_ret.loc[(quantile, 'mean')]
            mean.name = 'Quantile ' + str(quantile)
            mean.plot(ax=ax, color=palette[i])

            if std_bar:
                std = q_ret.loc[(quantile, 'std')]
                ax.errorbar(std.index, mean, yerr=std,
                            fmt='none', ecolor=palette[i], label='none')
            i += 1

        ax.axvline(x=0, color='k', linestyle='--')
        ax.legend()
        ax.set(ylabel='Mean Return (bps)',
               title=("Average Cumulative Returns by Quantile"
                      if title is None else title),
               xlabel='Periods')

    return ax


def plot_events_distribution(events, num_bars=50, ax=None):
    """
    Plots the distribution of events in time.

    Parameters
    ----------
    events : pd.Series
        A pd.Series whose index contains at least 'date' level.
    num_bars : integer, optional
        Number of bars to plot
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
    """

    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    start = events.index.get_level_values('date').min()
    end = events.index.get_level_values('date').max()
    group_interval = (end - start) / num_bars
    grouper = pd.Grouper(level='date', freq=group_interval)
    events.groupby(grouper).count().plot(kind="bar", grid=False, ax=ax)
    ax.set(ylabel='Number of events',
           title='Distribution of events in time',
           xlabel='Date')

    return ax


########################################################################################################
DECIMAL_TO_PCT = 100


def sd_plot_quantile_statistics_table(factor_data):
    quantile_stats = factor_data.groupby('factor_quantile') \
        .agg(['min', 'max', 'mean', 'std', 'count'])['factor']
    quantile_stats['daily_count'] = np.floor(quantile_stats['count']
                                             / len(factor_data.index.get_level_values('date').unique()))
    quantile_stats['count %'] = np.floor(quantile_stats['count']
                                         / quantile_stats['count'].sum() * 10000)/100

    print("FactorValues Quantiles Statistics")
    utils.print_table(quantile_stats)


def sd_plot_quantile_returns_bar(mean_ret_by_q,
                                 by_group=False,
                                 ylim_percentiles=None,
                                 ax=None):
    mean_ret_by_q = mean_ret_by_q.copy()

    if ylim_percentiles is not None:
        ymin = (np.nanpercentile(mean_ret_by_q.values,
                                 ylim_percentiles[0]) * DECIMAL_TO_PCT)
        ymax = (np.nanpercentile(mean_ret_by_q.values,
                                 ylim_percentiles[1]) * DECIMAL_TO_PCT)
    else:
        ymin = None
        ymax = None

    if by_group:
        num_group = len(
            mean_ret_by_q.index.get_level_values('group').unique())

        if ax is None:
            v_spaces = ((num_group - 1) // 2) + 1
            f, ax = plt.subplots(v_spaces, 2, sharex=False,
                                 sharey=True, figsize=(18, 6 * v_spaces))
            ax = ax.flatten()

        for a, (sc, cor) in zip(ax, mean_ret_by_q.groupby(level='group')):
            (cor.xs(sc, level='group')
                .multiply(DECIMAL_TO_PCT)
                .plot(kind='bar', title=sc, ax=a))

            a.set(xlabel='', ylabel='Mean Return (%)',
                  ylim=(ymin, ymax))

        if num_group < len(ax):
            ax[-1].set_visible(False)

        return ax

    else:
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(18, 6))

        (mean_ret_by_q.multiply(DECIMAL_TO_PCT)
            .plot(kind='bar',
                  title="Annualized Return By Factor Quantile", ax=ax))
        ax.set(xlabel='', ylabel='Annualized Return (%)',
               ylim=(ymin, ymax))

        return ax


def sd_plot_cumulative_returns_by_quantile(quantile_returns,
                                           period,
                                           longshort=True,
                                           ax=None):
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    ret_wide = quantile_returns.unstack('factor_quantile')

    if longshort:
        cum_ret = ret_wide.cumsum()
        cum_ret = cum_ret * DECIMAL_TO_PCT
    else:
        cum_ret = ret_wide.apply(perf.cumulative_returns)

    cum_ret = cum_ret.loc[:, ::-1]  # we want negative quantiles as 'red'

    cum_ret.plot(lw=2, ax=ax, cmap=cm.coolwarm)
    ax.legend()
    ymin, ymax = cum_ret.min().min(), cum_ret.max().max()
    ax.set(ylabel=('Cumulative Returns %' if longshort else 'Cumulative NetValue'),
           title=('''LongShort Cumulative Return by Quantile
                    ({} Period Forward Return)'''.format(period)
                  if longshort else
                  '''LongOnly Cumulative Return by Quantile
                           ({} Period Forward Return)'''.format(period)),
           xlabel='',
           #          yscale='linear' if longshort else 'symlog',
           yticks=np.linspace(ymin, ymax, 5),
           ylim=(ymin, ymax))

    ax.yaxis.set_major_formatter(ScalarFormatter())
    ax.axhline(1.0, linestyle='-', color='black', lw=1)

    return ax


def sd_plot_cumulative_returns_spread(factor_returns,
                                      period,
                                      title=None,
                                      ax=None):
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
        factor_returns = factor_returns.cumsum()

        factor_returns.plot(ax=ax, lw=3, alpha=0.6)
        ax.set(ylabel='Cumulative Returns',
               title=("Spread Cumulative Return and ReturnAE ({} Fwd Period)".format(period)
                      if title is None else title),
               xlabel='')
    return ax

def sd_plot_cumulative_excessreturn_by_quantile(quantile_returns,
                                                   period,
                                                   ax=None):
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))

    ret_wide = quantile_returns.unstack('factor_quantile')

    cum_ret = ret_wide.cumsum()
    cum_ret = cum_ret * DECIMAL_TO_PCT

    cum_ret = cum_ret.loc[:, ::-1]  # we want negative quantiles as 'red'

    cum_ret.plot(lw=2, ax=ax, cmap=cm.coolwarm)
    ax.legend()
    ymin, ymax = cum_ret.min().min(), cum_ret.max().max()
    ax.set(ylabel=('Cumulative ExcessReturns %'),
           title=('''Cumulative ExcessReturns by Quantile
                    ({} Period Forward Return)'''.format(period)
                ),
           xlabel='',
           #          yscale='linear' if longshort else 'symlog',
           yticks=np.linspace(ymin, ymax, 5),
           ylim=(ymin, ymax))

    ax.yaxis.set_major_formatter(ScalarFormatter())
    ax.axhline(1.0, linestyle='-', color='black', lw=1)

    return ax

def sd_plot_cumulative_excessreturn(factor_returns,
                                    period,
                                    title=None,
                                    ax=None):
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
        factor_returns = factor_returns.cumsum()

        factor_returns.plot(ax=ax, lw=3, alpha=0.6)
        ax.set(ylabel='Cumulative Returns',
               title=("Cumulative ExcessReturn and ReturnAE ({} Fwd Period)".format(period)
                      if title is None else title),
               xlabel='')
    return ax


import numpy as np
import pandas as pd
from pandas import Series,DataFrame
from typing import Iterable


def drop_duplicated_index(a:pd.Series):
    """去除重复索引的行"""
    return a[~a.index.duplicated()]

def drop_highcorr(df:DataFrame,gap=0.7):
    """
    删除高相关性因子，因子名为columns
    
    Return: 剩余因子名列表
    """
    corr = df.corr()
    while True:
        l = len(corr)
        index = corr.index
        for i in range(l-1):
            for j in range(i+1,l):
                loc = index[i]
                col = index[j]
                # print((i,j,loc,col))
                v = corr[col].loc[loc]
                if v > gap:
                    corr = corr.drop(index=[loc],columns=[loc])
                    break
            if len(corr) != l:
                break
        if len(corr) == l:
            break
    return list(corr.index)


class DataHandle:

    @staticmethod
    def handle_nan_value(c,method='drop'):
        """c: Series|Dataframe"""
        l = len(c)
        if method == 'drop':
            c = c.dropna()
            if len(c)/l < 0.5:
                print('You have too many None value, please check your data.')
        
        elif method == 'ffill':
            c = c.ffill().dropna()

        elif method == 'bfill':
            c = c.bfill().dropna()
        
        elif method == 'mfill':
            c = c.fillna(c.mean())
        
        elif method == 'fill0':
            c = c.fillna(0)
        
        else:
            raise ValueError('method should be drop, ffill, bfill, mfill, fill0.')
        
        return [c[col] for col in c.columns]

    @staticmethod
    def handle_nan_value_series(data:Iterable[Series],method='drop'):
        """
        This function is to handle the nan value and align index for two Series data with similar index, you can pick the different methed to deal with.

        Params:
        - data:Iterable[Series], you should input two Seriers at least.

        - method:
            drop: drop the nan value.
            ffill: ffill the nan value.
            bfill: bfill the nan value.
            mfill: use the mean value to fill the nan value.
            fill0: use 0 to fill the nan value.
        
        - Return: Iterable[Series]
        """
        c:DataFrame = pd.concat(data,axis=1)
        l = len(c)
        if method == 'drop':
            c = c.dropna()
            if len(c)/l < 0.5:
                print('You have too many None value, please check your data.')
        
        elif method == 'ffill':
            c = c.ffill().dropna()

        elif method == 'bfill':
            c = c.bfill().dropna()
        
        elif method == 'mfill':
            c = c.fillna(c.mean())
        
        elif method == 'fill0':
            c = c.fillna(0)
        
        else:
            raise ValueError('method should be drop, ffill, bfill, mfill, fill0.')
        
        return [c[col] for col in c.columns]
        

    @staticmethod
    def handle_nan_value_X_y(X:DataFrame,y:Series,method='drop'):
        """
        This function is to handle the nan value and align index for X and y, you can pick the different methed to deal with.

        Params:
        - X:DataFrame
        - y:Series

        - method: pick a method to deal with the X data.
            drop: drop the nan value.
            ffill: ffill the nan value.
            bfill: bfill the nan value.
            mfill: use the mean value to fill the nan value.
            fill0: use 0 to fill the nan value.
        
        Return: (X_,y_)

        Warning: use 'y' as the columns name, you should ensure 'y' is not in X.columns.
        """

        if 'y' in X.columns:
            raise ValueError(' str(y) should not in X.columns')
        
        l = len(X)

        if method == 'drop':
            X_ = X.dropna().copy()
            if len(X_)/l < 0.5:
                print('You have too many None value, please check your data.')
        
        elif method == 'ffill':
            X_ = X.ffill().dropna().copy()

        elif method == 'bfill':
            X_ = X.bfill().dropna().copy()
        
        elif method == 'mfill':
            X_ = X.fillna(X.mean()).copy()
        
        elif method == 'fill0':
            X_ = X.fillna(0).copy()
        
        else:
            raise ValueError('method should be drop, ffill, bfill, mfill, fill0.')
        
        y_ = y.copy()
        y_.name = 'y'
        data = pd.concat([y_,X_],axis=1).dropna()
        y_ = data['y']
        X_ = data.drop(columns=['y'])
        
        return (X_,y_)
    
    @staticmethod
    def meanize(data:DataFrame):
        return data - data.mean(numeric_only=True)
    
    @staticmethod
    def normalize(data:DataFrame):
        return (data - data.mean(numeric_only=True))/data.std(numeric_only=True)
    
    @staticmethod
    def shrink_extreme_value(data: pd.DataFrame):
        rate = 0.05
        lower = data.quantile(rate)
        upper = data.quantile(1 - rate)

        if isinstance(data, pd.DataFrame):
            axis_value = 1
        elif isinstance(data, pd.Series):
            axis_value = 0
        else:
            raise ValueError("Unsupported data type. Use DataFrame or Series.")

        return data.clip(lower=lower, upper=upper, axis=axis_value)
    
    @staticmethod
    def normalize_and_shrink(data:DataFrame):
        nor = DataHandle.normalize(data)
        nor_shr = DataHandle.shrink_extreme_value(nor)
        return nor_shr.values
    
    @staticmethod
    def multi_index_meanize(data:DataFrame,level=[0]):
        return data.groupby(level=level).transform(DataHandle.meanize)
    
    @staticmethod
    def multi_index_normalize(data:DataFrame,level=[0]):
        return data.groupby(level=level).transform(DataHandle.normalize)
    
    @staticmethod
    def multi_index_shrink(data:DataFrame,level=[0]):
        return data.groupby(level=level).transform(DataHandle.shrink_extreme_value)
    
    @staticmethod
    def group_normalize(data:DataFrame, group:dict):
        return None
        mean = data.groupby(by=group).mean(numeric_only=True)
        std = data.groupby(by=group).std(numeric_only=True)
        return (data-mean)/std
    
    @staticmethod
    def linear_wash(factor:DataFrame, q=0.8, rate=None, err=500):
        """
        factor: pd.DataFrame, 通过除以每列abs0.8分位点的值做线性变换, 若变换后的值仍大于10则视为异常值,替换为空缺值
        """
        l = len(factor)
        if rate is None:
            
            q = np.abs(factor).quantile(q)
        else:
            q = np.abs(factor.iloc[:int(l*rate)]).quantile(q)
        lf = factor/q
        linear_wash_factor = DataFrame(np.where(np.abs(lf)>err,np.nan,lf),index=lf.index,columns=lf.columns)
        return linear_wash_factor
    

#######################################################################################################################################################################################################


def order_array(value:np.ndarray):
    v = np.array(value)
    if len(v.shape) > 1:
        raise ValueError
    l = len(v)
    order_lst = [0 for i in range(l)]
    argsort = np.argsort(v)
    for i in range(l):
        order_lst[argsort[i]] = i 
    return np.array(order_lst)

def dataframe_to_order(df:pd.DataFrame):
    result = []
    isna = np.where(df.isna(),np.nan,1)
    for i in range(len(df)):
        v = df.iloc[i].values
        order_v = order_array(v)
        result.append(order_v)
    result = np.vstack(result)*isna
    return DataFrame(result,index=df.index,columns=df.columns)
    


def value_to_sort_rate(amt:pd.DataFrame,lower=0.5,upper=1,penalty=2):
    """按照每行的值，变成序并压缩为处于lower->upper的rate, penalty为惩罚系数,即指数上的数值,默认为平方项，随着penalty增加，对小序的惩罚增加
    
    -math: l + (u-l)*(argsort/max(argsort))**p
    
    """
    amt_order = dataframe_to_order(amt)
    return lower + (upper-lower)*amt_order.divide(amt_order.max(axis=1),axis=0)**2


import numpy as np
import pandas as pd
from pandas import Series,DataFrame
import tushare as ts
from joblib import Memory
import joblib
from datetime import datetime,timedelta
import os

current_file_path = os.path.abspath(__file__)
current_dir = os.path.dirname(current_file_path)
resources_dir = os.path.join(current_dir, 'resources')
calendar_file_path = os.path.join(resources_dir, 'stock.calendar')
basic_file_path = os.path.join(resources_dir, 'stock.basic')



STARTDATE = '19910101'
ENDDATE = '20991231'
def reload():   
    ts.set_token(token='afc6021c1b2e8029eed7f41e1a7f9f98b87bc84cdc9ee3f5755df06a')
    pro = ts.pro_api()

    def CALENDAR_trade_get():
        df = pro.trade_cal(exchange='',start_date=STARTDATE,end_date=ENDDATE)
        df.columns = ['exchange','cal_date','is_open','pretrade_date']
        df.set_index('cal_date',inplace=True)
        df.index = pd.to_datetime(df.index)
        df['pretrade_date'] = pd.to_datetime(df['pretrade_date'])
        df.sort_index(inplace=True)
        return df

    def stock_basic():
        data1 = pro.stock_basic(exchange='', list_status='L', fields='ts_code,symbol,name,area,industry,list_date,list_status,delist_date')
        data2 = pro.stock_basic(exchange='', list_status='D', fields='ts_code,symbol,name,area,industry,list_date,list_status,delist_date')
        data = pd.concat([data1,data2]).set_index(['symbol'])
        data['list_date'] = pd.to_datetime(data['list_date'])
        data['delist_date'] = pd.to_datetime(data['delist_date'])
        return data
    
    calendar_file_path = os.path.join(resources_dir, 'stock.calendar')
    
    CALENDAR = CALENDAR_trade_get()
    STOCKBASIC = stock_basic()
    joblib.dump(CALENDAR,calendar_file_path)
    joblib.dump(STOCKBASIC,basic_file_path)
    print("Reload calendar, stockbasic: Success!")

memory = Memory(location='./cache')

STARTDATE = '19910101'
ENDDATE = '20251231'


# @memory.cache
# def CALENDAR_trade_get():
#     df = pro.trade_cal(exchange='',start_date=STARTDATE,end_date=ENDDATE)
#     df.columns = ['exchange','cal_date','is_open','pretrade_date']
#     df.set_index('cal_date',inplace=True)
#     df.index = pd.to_datetime(df.index)
#     df['pretrade_date'] = pd.to_datetime(df['pretrade_date'])
#     df.sort_index(inplace=True)
#     return df

# @memory.cache
# def stock_basic():
#     data1 = pro.stock_basic(exchange='', list_status='L', fields='ts_code,symbol,name,area,industry,list_date,list_status,delist_date')
#     data2 = pro.stock_basic(exchange='', list_status='D', fields='ts_code,symbol,name,area,industry,list_date,list_status,delist_date')
#     data = pd.concat([data1,data2]).set_index(['symbol'])
#     data['list_date'] = pd.to_datetime(data['list_date'])
#     data['delist_date'] = pd.to_datetime(data['delist_date'])
#     return data


def unique(lst):
    uni = np.unique(np.array(lst))
    uni.sort()
    return uni

# CALENDAR = CALENDAR_trade_get()
# STOCKBASIC = stock_basic()

CALENDAR = joblib.load(calendar_file_path)
STOCKBASIC = joblib.load(basic_file_path)

class CALENDAR:
    
    def __init__(self,start_date = STARTDATE,end_date = ENDDATE, CALENDAR = CALENDAR):
        self.start_date = pd.to_datetime(start_date)
        self.end_date = pd.to_datetime(end_date)
        self.CALENDAR = CALENDAR
        
    def is_trade_date(self,date):
        if self.CALENDAR.loc[date]['is_open'] == 1:
            return True
        return False
    
    def trade_date_in_range(self,start_date=None,end_date=None,keeplast=True):
        '''get trade date between start_date and end_date
        if keeplast=False: 若end_date为交易日,则丢弃该日期,主要是防止分区间提取数据时发生日期重复问题
        '''
        tar = self.CALENDAR.loc[start_date:end_date].copy()
        tar = tar[tar['is_open']==1].index
        if keeplast:
            return tar
        if self.is_trade_date(pd.to_datetime(end_date)):
            return list(tar[:-1])
        else:
            return list(tar)
    
    def single_shift(self,date:datetime,days=0):
        '''get the last most recent trading date around date minus days'''
        assert isinstance(date,datetime), "input should be datetime type"
        date = date-timedelta(days)
        if date < self.start_date or date > self.end_date:
            return np.nan
        try:
            return self.trade_date_in_range(end_date=date)[-1]
        except:
            return np.nan 
        
    def single_next(self,date:datetime,days=0):
        '''get the first most recent trading date around date plus days'''
        assert isinstance(date,datetime), "input should be datetime type"
        date = date+timedelta(days)
        while date >= self.start_date or date <= self.end_date:
            if self.CALENDAR.loc[date]['is_open'] == 1:
                return date
            date += timedelta(1)
        return np.nan

    def shift(self,date_lst,days=0):
        '''get the last most recent trading dates around dates minus days'''
        if days==0:
            return pd.to_datetime(self.CALENDAR.loc[date_lst]['pretrade_date'].values)
        shift_date = []
        for date in date_lst:
            shift_date.append(self.single_shift(date,days=days))
        return pd.to_datetime(shift_date)
    
    def next(self,date_lst,days=0):
        '''get the next trading dates if current date is not a trading day'''
        if days==0:
            return pd.to_datetime(self.CALENDAR.loc[date_lst]['pretrade_date'].values)
        next_date = []
        for date in date_lst:
            next_date.append(self.single_next(date,days=days))
        return pd.to_datetime(next_date)
        
        
    def get_trade_date(self,date_lst):
        '''get the trade dates in the date list'''
        tar = self.CALENDAR.loc[date_lst].copy()
        return tar[tar['is_open']==1].index
    
    def monthly(self,start_date,end_date,loc=1):
        '''get the last trading day of each month if loc==1 or the first if loc==0 '''
        if loc == 1:
            date_lst = pd.date_range(start=start_date,end=end_date,freq='M')
            return self.shift(date_lst,days=0)
        if loc == 0:
            date_lst = pd.date_range(start=start_date,end=end_date,freq='MS')
            return self.next(date_lst,days=0)
        raise ValueError('param loc should be 0 or 1')
    
    def weekly(self,start_date,end_date,loc=1):
        '''get the last trading day of each week if loc==1 or the first if loc==0 '''
        if loc == 1:
            date_lst = pd.date_range(start=start_date,end=end_date,freq='W')
            return self.shift(date_lst,days=0).drop_duplicates().sort_values()
        if loc == 0:
            date_lst = pd.date_range(start=start_date,end=end_date,freq='WS')
            return self.next(date_lst,days=0).drop_duplicates().sort_values()
        raise ValueError('param loc should be 0 or 1')

CALENDAR_TOOL = CALENDAR()



from pandas import DataFrame
import pandas as pd
from .cta_alpha_factor import AlphaLC
from joblib import delayed,Parallel
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def trans_multicolumns(data):
    """data:pd.DataFrame, multiindex, level_0 assets, level_1 date, columns projects"""
    
    return data.unstack(level=0)
    # multi_col_data = []
    # for col in data.columns:
    #     d = data[col].unstack().reset_index()
    #     d['project'] = col
    #     multi_col_data.append(d)
    # multi_col_data = pd.concat(multi_col_data).set_index(['project','asset'])
    # return multi_col_data.T.sort_index()
    

class DailyMoveFactor:
    """Generate move factor based on daily bar.
    """
    def __init__(self, daily:DataFrame,min_length = 60):
        """daily:DataFrame. With datetime index, multicolumns and level_0 includes open,high,low,close,volume,amount,turnover, level_1 with assets."""
        if len(daily) < min_length:
            raise ValueError("len(daily) < min_length")
        self.daily = daily.sort_index()
        self.alpha = AlphaLC(df_data=self.daily)
        self.factor_name = [method for method in dir(AlphaLC) if callable(getattr(AlphaLC, method)) and not method.startswith("__")]
        
    def gen_factor(self,factor_name=None):
        if factor_name is None:
            factor_name = self.factor_name
            
        factor_data = []
        used_fac = []
        for name in factor_name:
            if hasattr(AlphaLC, name) and callable(getattr(AlphaLC, name)):
                try:
                    print("计算中：{}".format(name))
                    fac = getattr(self.alpha, name)()
                    fac.index = pd.to_datetime(fac.index)
                    fac = fac.stack(dropna=False).sort_index().astype('float32')
                    used_fac.append(name)
                    factor_data.append(fac)
                except:
                    print(f"函数 {name} 计算时出错")
                    continue
            else:
                print(f"函数 {name} 未找到")
        factor_data = pd.concat(factor_data,axis=1).sort_index()
        factor_data.columns = used_fac
        return factor_data.astype('float32')
    
    def quick_gen_factor(self,factor_name=None,n_job=8):
        if factor_name is None:
            factor_name = self.factor_name 
            
        def work_func(name):
            if hasattr(AlphaLC, name) and callable(getattr(AlphaLC, name)):
                try:
                    print("计算中：{}".format(name))
                    fac = getattr(self.alpha, name)()
                    fac.index = pd.to_datetime(fac.index)
                    fac = fac.stack(future_stack=True).sort_index().astype('float32')
                    fac.name = name
                    return fac
                except:
                    print(f"函数 {name} 计算时出错")
                    return None
            else:
                print(f"函数 {name} 未找到")
                return None
        
        results = Parallel(n_jobs=n_job)(delayed(work_func)(name) for name in factor_name)
        return results
        
        
av_factor = ['basic_in5','basic_in6','basic_in14','mean_in1','mean_in2','mean_in3','mean_in4','mean_in5','']            
        
        
        
        


import torch.nn as nn
import torch.nn.functional as F
import torch

import numpy as np
import pandas as pd
from pandas import Series,DataFrame

from torch.utils.data import TensorDataset, DataLoader

from math import log

from ..util import DLmodel
from ...stockcalendar import CALENDAR_TOOL


trade_dates = CALENDAR_TOOL.trade_date_in_range(start_date='20010101',end_date='20241201').copy(deep=True)




class BlockConvBase(nn.Module):
    
    def __init__(self, in_channels, out_channels, kernel_size):
        super(BlockConvBase,self).__init__()
        self.conv = nn.Conv2d(in_channels,out_channels,kernel_size)
    def forward(self,x):
        x = self.conv(x).squeeze(3).permute(0,2,1)
        return x


class BlockConvMax(nn.Module):
    
    def __init__(self,in_channels,out_channels,kernel_size) -> None:
        super(BlockConvMax,self).__init__()
        
        self.conv = nn.Conv2d(in_channels,out_channels,kernel_size)
        self.maxpool = nn.MaxPool2d((kernel_size[0],1),stride=1)
        # self.conv3 = nn.Conv2d(10,20,kernel_size=(5,5),padding=2)
        
    def forward(self,x):
        x = self.conv(x)
        x = self.maxpool(x).squeeze(3).permute(0,2,1)
        return x
    

class BlockConvAve(nn.Module):
    
    def __init__(self,in_channels,out_channels,kernel_size) -> None:
        super(BlockConvAve,self).__init__()
        
        self.conv = nn.Conv2d(in_channels,out_channels,kernel_size)
        self.avepool = nn.AvgPool2d((kernel_size[0],1),stride=1)
        # self.conv3 = nn.Conv2d(10,20,kernel_size=(5,5),padding=2)
        
    def forward(self,x):
        x = self.conv(x)
        x = self.avepool(x).squeeze(3).permute(0,2,1)
        return x
    


class BlockConvCross(nn.Module):
    
    def __init__(self,in_channels,out_channels,kernel_size) -> None:
        super(BlockConvCross,self).__init__()
        
        self.conv1 = nn.Conv2d(in_channels,out_channels,kernel_size)
        self.conv2 = nn.Conv2d(out_channels,out_channels,kernel_size)
        self.avepool1 = nn.AvgPool2d((kernel_size[0],1),stride=1)
        self.avepool2 = nn.AvgPool2d((kernel_size[0],1),stride=1)
        
    def forward(self,x):
        x = self.conv1(x)
        x = self.avepool1(x)
        x = self.conv2(x)
        x = self.avepool2(x)
        x = x.permute(0,2,1,3).reshape(x.shape[0],x.shape[2],-1)
        return x
 
class ConvFeatureGet0(nn.Module):
    
    def __init__(self, base_seq_len, base_dim, out_seq_len):
        super(ConvFeatureGet0,self).__init__()
        if base_seq_len < 140:
            raise ValueError("the seq_len should be >= 140 !!!!")
        self.out_seq_len = out_seq_len
        self.base_dim = base_dim
        
        self.conv_base1 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(1,self.base_dim))
        self.conv_base2 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(2,self.base_dim))
        self.conv_base3 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(3,self.base_dim))
        self.conv_base4 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(5,self.base_dim))
        self.conv_base5 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(8,self.base_dim))
        self.conv_base6 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(13,self.base_dim))
        self.conv_base7 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(21,self.base_dim))
        self.conv_base8 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(34,self.base_dim))
        self.conv_base9 = BlockConvBase(in_channels=1,out_channels=3,kernel_size=(55,self.base_dim))
        
        self.max_1 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(5,self.base_dim))
        self.max_2 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(8,self.base_dim))
        self.max_3 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(13,self.base_dim))
        self.max_4 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(21,self.base_dim))
        self.max_5 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(34,self.base_dim))
        
        self.ave_1 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(5,self.base_dim))
        self.ave_2 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(8,self.base_dim))
        self.ave_3 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(13,self.base_dim))
        self.ave_4 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(21,self.base_dim))
        self.ave_5 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(34,self.base_dim))
        
        self.cross_1 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(3,3))
        self.cross_2 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(5,5))
        self.cross_3 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(7,5))
        self.cross_4 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(11,5))
        self.cross_5 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(19,5))
        self.cross_6 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(21,7))
        
    def forward(self,x:torch.tensor):
        ### x: batch_size, seq_len, base_dim
        x = x.unsqueeze(1)
        base = self.base_block(x)
        max = self.max_block(x)
        ave = self.ave_block(x)
        # cross = self.cross_block(x)
        y = x.squeeze(1)[:,-self.out_seq_len:,:]
        return torch.cat([y,base,max,ave],dim=2)
    
    def base_block(self,x):
        
        base1 = self.conv_base1(x)[:,-self.out_seq_len:,:]
        base2 = self.conv_base2(x)[:,-self.out_seq_len:,:]
        base3 = self.conv_base3(x)[:,-self.out_seq_len:,:]
        base4 = self.conv_base4(x)[:,-self.out_seq_len:,:]
        base5 = self.conv_base5(x)[:,-self.out_seq_len:,:]
        base6 = self.conv_base6(x)[:,-self.out_seq_len:,:]
        base7 = self.conv_base7(x)[:,-self.out_seq_len:,:]
        base8 = self.conv_base8(x)[:,-self.out_seq_len:,:]
        base9 = self.conv_base9(x)[:,-self.out_seq_len:,:]
        # return torch.cat([base1,base2,base3],dim=2)
        return torch.cat([base1,base2,base3,base4,base5,base6,base7,base8,base9],dim=2)
    
    def max_block(self,x):
        
        max1 = self.max_1(x)[:,-self.out_seq_len:,:]
        max2 = self.max_2(x)[:,-self.out_seq_len:,:]
        max3 = self.max_3(x)[:,-self.out_seq_len:,:]
        max4 = self.max_4(x)[:,-self.out_seq_len:,:]
        # max5 = self.max_5(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([max1,max2,max3,max4],dim=2)
    
    def ave_block(self,x):
        
        ave1 = self.ave_1(x)[:,-self.out_seq_len:,:]
        ave2 = self.ave_2(x)[:,-self.out_seq_len:,:]
        ave3 = self.ave_3(x)[:,-self.out_seq_len:,:]
        ave4 = self.ave_4(x)[:,-self.out_seq_len:,:]
        # ave5 = self.ave_5(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([ave1,ave2,ave3,ave4],dim=2)
    
    def cross_block(self,x):
        # cross1 = self.cross_1(x)[:,-self.out_seq_len:,:]
        # cross2 = self.cross_2(x)[:,-self.out_seq_len:,:]
        # cross3 = self.cross_3(x)[:,-self.out_seq_len:,:]
        cross4 = self.cross_4(x)[:,-self.out_seq_len:,:]
        cross5 = self.cross_5(x)[:,-self.out_seq_len:,:]
        cross6 = self.cross_6(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([cross4,cross5,cross6],dim=2)

   
    
class ConvFeatureGet1(nn.Module):
    
    def __init__(self, base_seq_len, base_dim, out_seq_len):
        super(ConvFeatureGet1,self).__init__()
        if base_seq_len < 140:
            raise ValueError("the seq_len should be >= 140 !!!!")
        self.out_seq_len = out_seq_len
        self.base_dim = base_dim
        
        self.conv_base1 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(1,self.base_dim))
        self.conv_base2 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(2,self.base_dim))
        self.conv_base3 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(3,self.base_dim))
        self.conv_base4 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(5,self.base_dim))
        self.conv_base5 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(8,self.base_dim))
        self.conv_base6 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(13,self.base_dim))
        self.conv_base7 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(21,self.base_dim))
        self.conv_base8 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(34,self.base_dim))
        self.conv_base9 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(55,self.base_dim))
        
        self.max_1 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(5,self.base_dim))
        self.max_2 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(8,self.base_dim))
        self.max_3 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(13,self.base_dim))
        self.max_4 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(21,self.base_dim))
        self.max_5 = BlockConvMax(in_channels=1,out_channels=3,kernel_size=(34,self.base_dim))
        
        self.ave_1 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(5,self.base_dim))
        self.ave_2 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(8,self.base_dim))
        self.ave_3 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(13,self.base_dim))
        self.ave_4 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(21,self.base_dim))
        self.ave_5 = BlockConvAve(in_channels=1,out_channels=3,kernel_size=(34,self.base_dim))
        
        self.cross_1 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(3,3))
        self.cross_2 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(5,5))
        self.cross_3 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(7,5))
        self.cross_4 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(11,5))
        self.cross_5 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(19,5))
        self.cross_6 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(21,7))
        
    def forward(self,x:torch.tensor):
        ### x: batch_size, seq_len, base_dim
        x = x.unsqueeze(1)
        base = self.base_block(x)
        max = self.max_block(x)
        ave = self.ave_block(x)
        # cross = self.cross_block(x)
        y = x.squeeze(1)[:,-self.out_seq_len:,:]
        return torch.cat([y,base,max,ave],dim=2)
    
    def base_block(self,x):
        
        base1 = self.conv_base1(x)[:,-self.out_seq_len:,:]
        base2 = self.conv_base2(x)[:,-self.out_seq_len:,:]
        base3 = self.conv_base3(x)[:,-self.out_seq_len:,:]
        base4 = self.conv_base4(x)[:,-self.out_seq_len:,:]
        base5 = self.conv_base5(x)[:,-self.out_seq_len:,:]
        base6 = self.conv_base6(x)[:,-self.out_seq_len:,:]
        base7 = self.conv_base7(x)[:,-self.out_seq_len:,:]
        base8 = self.conv_base8(x)[:,-self.out_seq_len:,:]
        base9 = self.conv_base9(x)[:,-self.out_seq_len:,:]
        # return torch.cat([base1,base2,base3],dim=2)
        return torch.cat([base1,base2,base3,base4,base5,base6,base7,base8,base9],dim=2)
    
    def max_block(self,x):
        
        max1 = self.max_1(x)[:,-self.out_seq_len:,:]
        max2 = self.max_2(x)[:,-self.out_seq_len:,:]
        max3 = self.max_3(x)[:,-self.out_seq_len:,:]
        max4 = self.max_4(x)[:,-self.out_seq_len:,:]
        max5 = self.max_5(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([max1,max2,max3,max4,max5],dim=2)
    
    def ave_block(self,x):
        
        ave1 = self.ave_1(x)[:,-self.out_seq_len:,:]
        ave2 = self.ave_2(x)[:,-self.out_seq_len:,:]
        ave3 = self.ave_3(x)[:,-self.out_seq_len:,:]
        ave4 = self.ave_4(x)[:,-self.out_seq_len:,:]
        ave5 = self.ave_5(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([ave1,ave2,ave3,ave4,ave5],dim=2)
    
    def cross_block(self,x):
        # cross1 = self.cross_1(x)[:,-self.out_seq_len:,:]
        # cross2 = self.cross_2(x)[:,-self.out_seq_len:,:]
        # cross3 = self.cross_3(x)[:,-self.out_seq_len:,:]
        cross4 = self.cross_4(x)[:,-self.out_seq_len:,:]
        cross5 = self.cross_5(x)[:,-self.out_seq_len:,:]
        cross6 = self.cross_6(x)[:,-self.out_seq_len:,:]
        
        return torch.cat([cross4,cross5,cross6],dim=2)
        
class ConvFeatureGet2(nn.Module):
    
    def __init__(self, base_seq_len, base_dim, out_seq_len):
        super(ConvFeatureGet2,self).__init__()
        if base_seq_len < 120:
            raise ValueError("the seq_len should be >= 140 !!!!")
        self.out_seq_len = out_seq_len
        self.base_dim = base_dim
        
        self.conv_base3 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(3,self.base_dim))
        self.conv_base4 = BlockConvBase(in_channels=1,out_channels=5,kernel_size=(5,self.base_dim))
        self.conv_base5 = BlockConvBase(in_channels=1,out_channels=7,kernel_size=(8,self.base_dim))
        self.conv_base6 = BlockConvBase(in_channels=1,out_channels=10,kernel_size=(13,self.base_dim))

        
        self.max_1 = BlockConvMax(in_channels=1,out_channels=5,kernel_size=(5,self.base_dim))
        self.max_2 = BlockConvMax(in_channels=1,out_channels=7,kernel_size=(8,self.base_dim))
        self.max_3 = BlockConvMax(in_channels=1,out_channels=10,kernel_size=(13,self.base_dim))

        
        self.ave_1 = BlockConvAve(in_channels=1,out_channels=5,kernel_size=(5,self.base_dim))
        self.ave_2 = BlockConvAve(in_channels=1,out_channels=7,kernel_size=(8,self.base_dim))
        self.ave_3 = BlockConvAve(in_channels=1,out_channels=10,kernel_size=(13,self.base_dim))

        
        self.cross_1 = BlockConvCross(in_channels=1,out_channels=5,kernel_size=(3,5))
        self.cross_2 = BlockConvCross(in_channels=1,out_channels=7,kernel_size=(5,5))
        self.cross_3 = BlockConvCross(in_channels=1,out_channels=10,kernel_size=(7,5))
        self.cross_4 = BlockConvCross(in_channels=1,out_channels=10,kernel_size=(11,7))
        
    def forward(self,x:torch.tensor):
        ### x: batch_size, seq_len, base_dim
        x = x.unsqueeze(1)
        # base = self.base_block(x)
        max = self.max_block(x)
        ave = self.ave_block(x)
        # cross = self.cross_block(x)
        # y = x.squeeze(1)[:,-self.out_seq_len:,:]
        return torch.cat([max,ave],dim=2)
    
    def base_block(self,x):
        
        base3 = self.conv_base3(x)[:,-self.out_seq_len:,:]
        base4 = self.conv_base4(x)[:,-self.out_seq_len:,:]
        base5 = self.conv_base5(x)[:,-self.out_seq_len:,:]
        base6 = self.conv_base6(x)[:,-self.out_seq_len:,:]

        # return torch.cat([base1,base2,base3],dim=2)
        return torch.cat([base3,base4,base5,base6],dim=2)
    
    def max_block(self,x):
        
        max1 = self.max_1(x)[:,-self.out_seq_len:,:]
        max2 = self.max_2(x)[:,-self.out_seq_len:,:]
        max3 = self.max_3(x)[:,-self.out_seq_len:,:]

        return torch.cat([max1,max2,max3],dim=2)
    
    def ave_block(self,x):
        
        ave1 = self.ave_1(x)[:,-self.out_seq_len:,:]
        ave2 = self.ave_2(x)[:,-self.out_seq_len:,:]
        ave3 = self.ave_3(x)[:,-self.out_seq_len:,:]

        
        return torch.cat([ave1,ave2,ave3],dim=2)
    
    def cross_block(self,x):
        cross1 = self.cross_1(x)[:,-self.out_seq_len:,:]
        cross2 = self.cross_2(x)[:,-self.out_seq_len:,:]
        cross3 = self.cross_3(x)[:,-self.out_seq_len:,:]
        cross4 = self.cross_4(x)[:,-self.out_seq_len:,:]

        # return torch.cat([cross5,cross6],dim=2)
        return torch.cat([cross1,cross2,cross3,cross4],dim=2)        
        
class PositionalEncoding(nn.Module):
    """
    位置编码, 添加时要求维度为seq_len * batch_size * dim
    """
    def __init__(self, d_model, max_len=1000):
        super(PositionalEncoding, self).__init__()
        # 创建一个足够长的P
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return x


class CNNTransformer1(nn.Module):
    
    def __init__(self,base_seq_len,base_dim,seq_len,feature_dim,
                 num_heads,num_layers,d_model,output_dim,dim_feedforward,drop_rate,backward):
        super(CNNTransformer1,self).__init__()
        self.feature_dim = feature_dim
        self.d_model = d_model
        self.backward = backward
        
        
        self.pos_encoder = PositionalEncoding(d_model, max_len=1000)
        
        self.feature_get = ConvFeatureGet1(base_seq_len,base_dim,seq_len)
        
        self.pre_layer = nn.Linear(feature_dim,d_model)
        
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=drop_rate,
            activation='relu',
            batch_first=True,
        )
        
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        
        
        self.fc1 = nn.Linear(backward*self.d_model, 300)
        self.fc2 = nn.Linear(300,100)
        self.fc3 = nn.Linear(100,output_dim)
        
        self.aft_dropout1 = nn.Dropout(drop_rate)
        self.aft_dropout2 = nn.Dropout(drop_rate)

    def forward(self,x):
        
        x = self.feature_get(x)
        x = self.pre_layer(x)
        x = x.transpose(0,1)
        x = self.pos_encoder(x).transpose(0,1)
        
        x = self.encoder(x)
        
        
        out = x[:,-self.backward:,:].reshape(-1,self.backward*self.d_model)
        # 通过全连接层得到最终输出
        
        out = F.relu(self.fc1(out))
        out = self.aft_dropout1(out)
        out = F.relu(self.fc2(out))
        out = self.aft_dropout2(out)
        out = self.fc3(out)
        
        return out

class BlockCNNTransformer1(nn.Module):
    
    def __init__(self,base_seq_len,base_dim,seq_len,feature_dim,
                 num_heads,num_layers,d_model,dim_feedforward,drop_rate,backward):
        super(BlockCNNTransformer1,self).__init__()
        self.feature_dim = feature_dim
        self.d_model = d_model
        self.backward = backward
        
        
        self.pos_encoder = PositionalEncoding(d_model, max_len=1000)
        
        self.feature_get = ConvFeatureGet0(base_seq_len,base_dim,seq_len)
        
        self.pre_layer = nn.Linear(feature_dim,d_model)
        
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=drop_rate,
            activation='relu',
            batch_first=True,
        )
        
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        

    def forward(self,x):
        
        x = self.feature_get(x)
        x = self.pre_layer(x)
        x = x.transpose(0,1)
        x = self.pos_encoder(x).transpose(0,1)
        x = self.encoder(x)
        out = x[:,-self.backward:,:].reshape(-1,self.backward*self.d_model)
        
        return out


class CNNTransformer2(nn.Module):
    
    def __init__(self,base_seq_len,base_dim,seq_len,feature_dim,
                 num_heads,num_layers,d_model,output_dim,dim_feedforward,drop_rate,backward):
        super(CNNTransformer2,self).__init__()
        self.feature_dim = feature_dim
        self.d_model = d_model
        self.backward = backward
        
        
        self.pos_encoder = PositionalEncoding(d_model, max_len=1000)
        
        self.feature_get = ConvFeatureGet2(base_seq_len,base_dim,seq_len)
        
        self.pre_layer = nn.Linear(feature_dim,d_model)
        
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=drop_rate,
            activation='relu',
            batch_first=True,
        )
        
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        
        
        self.fc1 = nn.Linear(backward*self.d_model, 300)
        self.fc2 = nn.Linear(300,100)
        self.fc3 = nn.Linear(100,output_dim)
        
        self.aft_dropout1 = nn.Dropout(drop_rate)
        self.aft_dropout2 = nn.Dropout(drop_rate)

    def forward(self,x):
        
        x = self.feature_get(x)
        x = self.pre_layer(x)
        x = x.transpose(0,1)
        x = self.pos_encoder(x).transpose(0,1)
        
        x = self.encoder(x)
        
        
        out = x[:,-self.backward:,:].reshape(-1,self.backward*self.d_model)
        # 通过全连接层得到最终输出
        
        out = F.relu(self.fc1(out))
        out = self.aft_dropout1(out)
        out = F.relu(self.fc2(out))
        out = self.aft_dropout2(out)
        out = self.fc3(out)
        
        return out

# class BaseFactorTransformer(DLmodel):
    
#     def __init__(self, model_params: dict, class_model=CNNTransformer1, dtype='float') -> None:
#         super().__init__(model_params, class_model, dtype)
        
#     def reload_model(self,path):
#         self.model = torch.load(path)
        

#     def train(self,dataset,EPOCHS,batch_size=512,scale=100,shuffle=True,loss='Huber',lr=0.001,stop = 100):
        
#         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        
#         if loss[:5] == 'Huber':
#             if len(loss) == 5:
#                 delta = 1
#             else:
#                 delta = float(loss[5:])
#             loss_func = nn.HuberLoss(delta=delta)
#         elif loss == 'MSE':
#             loss_func = nn.MSELoss()
#         elif loss == 'L1':
#             loss_func = nn.L1Loss()
#         else:
#             loss_func = nn.MSELoss()
        
#         optim = torch.optim.Adam(self.model.parameters(),lr=lr)
#         loss_lst = []
#         r2score_lst = []
#         for epoch in range(EPOCHS):
#             self.model.train()
#             epoch_r2score = []
#             epoch_loss = []
#             for batch_idx, (data,target) in enumerate(dataloader):
#                 data = data.float().cuda()
#                 target = target.float().cuda()*scale
#                 optim.zero_grad()
#                 output = self.model(data)
#                 # print(output.shape)
#                 # print(target.shape)
                
#                 l = loss_func(output,target)
#                 if torch.isnan(l):
#                     print(data.isnan().sum())
#                     print(data.max())
#                     print(data.min())
#                     print(output.isnan().sum())
#                     print(output.max())
#                     print(output.min())
#                     print(target.isnan().sum())
#                     print(target.max())
#                     print(target.min())
                    
#                     return 0
    
#                 r2score = self.r2_score(target,output)[0].item()
#                 epoch_r2score.append(r2score)
#                 epoch_loss.append(l.item())
                
#                 l.backward()
                
#                 optim.step()
                
#                 if batch_idx % 100 == 0:
#                     print("train epoch:{}, batch_idx:{}, loss: {:.4f}, R2_score:{:.4f}".format(epoch+1,batch_idx,l.item(),r2score))
            
#             epoch_r2score = np.array(epoch_r2score)
#             epoch_loss = np.array(epoch_loss)
#             if epoch_r2score.mean() > stop:
#                 print("STOP: train epoch:{}, loss:{:.4f}, R2_score:{:.4f}".format(epoch+1,epoch_loss.mean(),epoch_r2score.mean()))
#                 print("")
#                 break
#             print("FINISH: train epoch:{}, loss:{:.4f}, R2_score:{:.4f}".format(epoch+1,epoch_loss.mean(),epoch_r2score.mean()))
#             print("")
     
#         self._r2score_lst = r2score_lst
#         self._loss_lst = loss_lst

#         return 1
    
#     @staticmethod
#     def gen_train_dataset(factor,price,dates,seq_len,forward,dropmid=None,dtype='float'):
#         seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
#         def single_point(asset,date,seq_len):
#             pre_date = seek_backward_dates.loc[date]
#             df = factor.loc[asset].loc[pre_date:date]
#             if len(df) < seq_len*0.9:
#                 return None
#             else:
#                 df = factor.loc[asset].loc[:date].iloc[-seq_len:].values
#                 if len(df) != seq_len:
#                     return None
#                 return ((asset,date),df)
        
#         X_ = []
#         y_ = []
        
#         index = []
        
#         def drop_mid_Series(sr:Series,dropmid):
#             sr = sr.sort_values()
#             l = len(sr)
#             if l < 100:
#                 return sr
#             return pd.concat([sr.iloc[:int(l*(1/2-dropmid/2))],sr.iloc[int(l*(1/2+dropmid/2)):]])
#         forward_returns = (price.shift(-forward)/price - 1)
#         for date in dates:
#             y:Series = forward_returns.loc[date].dropna()
#             if len(y) == 0:
#                 continue
#             if dropmid is not None:
#                 y = drop_mid_Series(y,dropmid)
#             for asset in y.index:
#                 dv = single_point(asset,date,seq_len)
#                 if dv is None:
#                     continue
                
#                 X_.append(dv[1])
#                 y_.append(y.loc[asset])
#                 index.append(dv[0])

#         if dtype == 'float':
#             X_ = torch.tensor(np.array(X_)).float()
#             y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1).float()
#         else:
#             X_ = torch.tensor(np.array(X_))
#             y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1)
            
#         index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
#         return (index,TensorDataset(X_,y_))
    
#     @staticmethod
#     def test(data:torch.tensor,index,model,batch_size=2048):
#         model = model.cuda()
#         result = []
#         with torch.no_grad():
#             model.eval()
#             for i in range(0,data.shape[0],batch_size):
#                 feature = data[i:i+batch_size].float().cuda()
#                 label = model(feature).reshape(-1).cpu().detach().numpy()
#                 result.append(label)
#         result = np.hstack(result)
#         factor = Series(result,index=index).sort_index()
#         return factor
    


                
                


from setuptools import setup, find_packages

setup(
    name='YsFinance',
    version='2.0.8',
    author='YunSheng',
    author_email='278211638@qq.com',
    description='About finance',
    # url='https://github.com/YunSheng0129/YsFinance.git',
    packages=find_packages(),
    package_data={
        'YsFinance': ['resources/info.derive.finance'],
    },
    install_requires=['pandas<=1.6',
                      'numpy<=1.26',
                      'scipy',
                      'tushare',
                      'scipy',
                      'statsmodels',
                      'joblib',
                      'ipykernel',
                      'empyrical'],  # 依赖列表
)

## python setup.py develop
## python setup.py sdist
## twine upload dist/*


#
# Copyright 2017 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from . import performance as perf
from . import plotting
from . import utils
import empyrical as ep
from scipy import stats

class GridFigure(object):
    """
    It makes life easier with grid plots
    """

    def __init__(self, rows, cols):
        self.rows = rows
        self.cols = cols
        self.fig = plt.figure(figsize=(14, rows * 7))
        self.gs = gridspec.GridSpec(rows, cols, wspace=0.4, hspace=0.3)
        self.curr_row = 0
        self.curr_col = 0

    def next_row(self):
        if self.curr_col != 0:
            self.curr_row += 1
            self.curr_col = 0
        subplt = plt.subplot(self.gs[self.curr_row, :])
        self.curr_row += 1
        return subplt

    def next_cell(self):
        if self.curr_col >= self.cols:
            self.curr_row += 1
            self.curr_col = 0
        subplt = plt.subplot(self.gs[self.curr_row, self.curr_col])
        self.curr_col += 1
        return subplt

    def close(self):
        plt.close(self.fig)
        self.fig = None
        self.gs = None


@plotting.customize
def create_summary_tear_sheet(
        factor_data, long_short=True, group_neutral=False
):
    """
    Creates a small summary tear sheet with returns, information, and turnover
    analysis.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    long_short : bool
        Should this computation happen on a long short portfolio? if so, then
        mean quantile returns will be demeaned across the factor universe.
    group_neutral : bool
        Should this computation happen on a group neutral portfolio? if so,
        returns demeaning will occur on the group level.
    """

    # Returns Analysis
    mean_quant_ret, std_quantile = perf.mean_return_by_quantile(
        factor_data,
        by_group=False,
        demeaned=long_short,
        group_adjust=group_neutral,
    )

    mean_quant_rateret = mean_quant_ret.apply(
        utils.rate_of_return, axis=0, base_period=mean_quant_ret.columns[0]
    )

    mean_quant_ret_bydate, std_quant_daily = perf.mean_return_by_quantile(
        factor_data,
        by_date=True,
        by_group=False,
        demeaned=long_short,
        group_adjust=group_neutral,
    )

    mean_quant_rateret_bydate = mean_quant_ret_bydate.apply(
        utils.rate_of_return,
        axis=0,
        base_period=mean_quant_ret_bydate.columns[0],
    )

    compstd_quant_daily = std_quant_daily.apply(
        utils.std_conversion, axis=0, base_period=std_quant_daily.columns[0]
    )

    alpha_beta = perf.factor_alpha_beta(
        factor_data, demeaned=long_short, group_adjust=group_neutral
    )

    mean_ret_spread_quant, std_spread_quant = perf.compute_mean_returns_spread(
        mean_quant_rateret_bydate,
        factor_data["factor_quantile"].max(),
        factor_data["factor_quantile"].min(),
        std_err=compstd_quant_daily,
    )

    periods = utils.get_forward_returns_columns(factor_data.columns)
    periods = list(map(lambda p: pd.Timedelta(p).days, periods))

    fr_cols = len(periods)
    vertical_sections = 2 + fr_cols * 3

    plotting.plot_quantile_statistics_table(factor_data)

    plotting.plot_returns_table(
        alpha_beta, mean_quant_rateret, mean_ret_spread_quant
    )

    plotting.plot_quantile_returns_bar(
        mean_quant_rateret,
        by_group=False,
        ylim_percentiles=None
    )

    # Information Analysis
    ic = perf.factor_information_coefficient(factor_data)
    plotting.plot_information_table(ic)

    # Turnover Analysis
    quantile_factor = factor_data["factor_quantile"]

    quantile_turnover = {
        p: pd.concat(
            [
                perf.quantile_turnover(quantile_factor, q, p)
                for q in range(1, int(quantile_factor.max()) + 1)
            ],
            axis=1,
        )
        for p in periods
    }

    autocorrelation = pd.concat(
        [
            perf.factor_rank_autocorrelation(factor_data, period)
            for period in periods
        ],
        axis=1,
    )

    plotting.plot_turnover_table(autocorrelation, quantile_turnover)

    plt.show()


@plotting.customize
def create_returns_tear_sheet(
        factor_data, long_short=True, group_neutral=False, by_group=False
):
    """
    Creates a tear sheet for returns analysis of a factor.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to,
        and (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    long_short : bool
        Should this computation happen on a long short portfolio? if so, then
        mean quantile returns will be demeaned across the factor universe.
        Additionally factor values will be demeaned across the factor universe
        when factor weighting the portfolio for cumulative returns plots
    group_neutral : bool
        Should this computation happen on a group neutral portfolio? if so,
        returns demeaning will occur on the group level.
        Additionally each group will weight the same in cumulative returns
        plots
    by_group : bool
        If True, display graphs separately for each group.
    """

    factor_returns = perf.factor_returns(
        factor_data, long_short, group_neutral
    )

    mean_quant_ret, std_quantile = perf.mean_return_by_quantile(
        factor_data,
        by_group=False,
        demeaned=long_short,
        group_adjust=group_neutral,
    )

    mean_quant_rateret = mean_quant_ret.apply(
        utils.rate_of_return, axis=0, base_period=mean_quant_ret.columns[0]
    )

    mean_quant_ret_bydate, std_quant_daily = perf.mean_return_by_quantile(
        factor_data,
        by_date=True,
        by_group=False,
        demeaned=long_short,
        group_adjust=group_neutral,
    )

    mean_quant_rateret_bydate = mean_quant_ret_bydate.apply(
        utils.rate_of_return,
        axis=0,
        base_period=mean_quant_ret_bydate.columns[0],
    )

    compstd_quant_daily = std_quant_daily.apply(
        utils.std_conversion, axis=0, base_period=std_quant_daily.columns[0]
    )

    alpha_beta = perf.factor_alpha_beta(
        factor_data, factor_returns, long_short, group_neutral
    )

    mean_ret_spread_quant, std_spread_quant = perf.compute_mean_returns_spread(
        mean_quant_rateret_bydate,
        factor_data["factor_quantile"].max(),
        factor_data["factor_quantile"].min(),
        std_err=compstd_quant_daily,
    )

    fr_cols = len(factor_returns.columns)
    vertical_sections = 2 + fr_cols * 3

    plotting.plot_returns_table(
        alpha_beta, mean_quant_rateret, mean_ret_spread_quant
    )

    plotting.plot_quantile_returns_bar(
        mean_quant_rateret,
        by_group=False,
        ylim_percentiles=None
    )

    plotting.plot_quantile_returns_violin(
        mean_quant_rateret_bydate, ylim_percentiles=(1, 99)
    )

    trading_calendar = factor_data.index.levels[0].freq
    if trading_calendar is None:
        trading_calendar = pd.tseries.offsets.BDay()
        warnings.warn(
            "'freq' not set in factor_data index: assuming business day",
            UserWarning,
        )

    # Compute cumulative returns from daily simple returns, if '1D'
    # returns are provided.
    if "1D" in factor_returns:
        title = (
            "Factor Weighted "
            + ("Group Neutral " if group_neutral else "")
            + ("Long/Short " if long_short else "")
            + "Portfolio Cumulative Return (1D Period)"
        )

        plotting.plot_cumulative_returns(
            factor_returns["1D"], period="1D", title=title
        )

        plotting.plot_cumulative_returns_by_quantile(
            mean_quant_ret_bydate["1D"], period="1D"
        )

    plotting.plot_mean_quantile_returns_spread_time_series(
        mean_ret_spread_quant,
        std_err=std_spread_quant,
        bandwidth=0.5
    )

    if by_group:
        (
            mean_return_quantile_group,
            mean_return_quantile_group_std_err,
        ) = perf.mean_return_by_quantile(
            factor_data,
            by_date=False,
            by_group=True,
            demeaned=long_short,
            group_adjust=group_neutral,
        )

        mean_quant_rateret_group = mean_return_quantile_group.apply(
            utils.rate_of_return,
            axis=0,
            base_period=mean_return_quantile_group.columns[0],
        )

        num_groups = len(
            mean_quant_rateret_group.index.get_level_values("group").unique()
        )

        vertical_sections = 1 + (((num_groups - 1) // 2) + 1)

        plotting.plot_quantile_returns_bar(
            mean_quant_rateret_group,
            by_group=True,
            ylim_percentiles=(5, 95)
        )


@plotting.customize
def create_information_tear_sheet(
        factor_data, group_neutral=False, by_group=False
):
    """
    Creates a tear sheet for information analysis of a factor.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    group_neutral : bool
        Demean forward returns by group before computing IC.
    by_group : bool
        If True, display graphs separately for each group.
    """

    ic = perf.factor_information_coefficient(factor_data, group_neutral)

    plotting.plot_information_table(ic)

    columns_wide = 2
    fr_cols = len(ic.columns)
    rows_when_wide = ((fr_cols - 1) // columns_wide) + 1
    vertical_sections = fr_cols + 3 * rows_when_wide + 2 * fr_cols

    plotting.plot_ic_ts(ic)

    plotting.plot_ic_hist(ic)
    plotting.plot_ic_qq(ic)

    if not by_group:
        mean_monthly_ic = perf.mean_information_coefficient(
            factor_data,
            group_adjust=group_neutral,
            by_group=False,
            by_time="M",
        )
        plotting.plot_monthly_ic_heatmap(
            mean_monthly_ic
        )

    if by_group:
        mean_group_ic = perf.mean_information_coefficient(
            factor_data, group_adjust=group_neutral, by_group=True
        )

        plotting.plot_ic_by_group(mean_group_ic)


@plotting.customize
def create_turnover_tear_sheet(factor_data, turnover_periods=None):
    """
    Creates a tear sheet for analyzing the turnover properties of a factor.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    turnover_periods : sequence[string], optional
        Periods to compute turnover analysis on. By default periods in
        'factor_data' are used but custom periods can provided instead. This
        can be useful when periods in 'factor_data' are not multiples of the
        frequency at which factor values are computed i.e. the periods
        are 2h and 4h and the factor is computed daily and so values like
        ['1D', '2D'] could be used instead
    """

    if turnover_periods is None:
        input_periods = utils.get_forward_returns_columns(
            factor_data.columns, require_exact_day_multiple=True,
        ).to_numpy()
        turnover_periods = utils.timedelta_strings_to_integers(input_periods)
    else:
        turnover_periods = utils.timedelta_strings_to_integers(
            turnover_periods,
        )

    quantile_factor = factor_data["factor_quantile"]

    quantile_turnover = {
        p: pd.concat(
            [
                perf.quantile_turnover(quantile_factor, q, p)
                for q in quantile_factor.sort_values().unique().tolist()
            ],
            axis=1,
        )
        for p in turnover_periods
    }

    autocorrelation = pd.concat(
        [
            perf.factor_rank_autocorrelation(factor_data, period)
            for period in turnover_periods
        ],
        axis=1,
    )

    plotting.plot_turnover_table(autocorrelation, quantile_turnover)

    fr_cols = len(turnover_periods)
    columns_wide = 1
    rows_when_wide = ((fr_cols - 1) // 1) + 1
    vertical_sections = fr_cols + 3 * rows_when_wide + 2 * fr_cols

    for period in turnover_periods:
        if quantile_turnover[period].isnull().all().all():
            continue
        plotting.plot_top_bottom_quantile_turnover(
            quantile_turnover[period], period=period
        )

    for period in autocorrelation:
        if autocorrelation[period].isnull().all():
            continue
        plotting.plot_factor_rank_auto_correlation(
            autocorrelation[period], period=period
        )


@plotting.customize
def create_full_tear_sheet(factor_data,
                           long_short=True,
                           group_neutral=False,
                           by_group=False):
    """
    Creates a full tear sheet for analysis and evaluating single
    return predicting (alpha) factor.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    long_short : bool
        Should this computation happen on a long short portfolio?
        - See tears.create_returns_tear_sheet for details on how this flag
        affects returns analysis
    group_neutral : bool
        Should this computation happen on a group neutral portfolio?
        - See tears.create_returns_tear_sheet for details on how this flag
        affects returns analysis
        - See tears.create_information_tear_sheet for details on how this
        flag affects information analysis
    by_group : bool
        If True, display graphs separately for each group.
    """

    plotting.plot_quantile_statistics_table(factor_data)
    create_returns_tear_sheet(
        factor_data, long_short, group_neutral, by_group, set_context=False
    )
    create_information_tear_sheet(
        factor_data, group_neutral, by_group, set_context=False
    )
    create_turnover_tear_sheet(factor_data, set_context=False)
    plotting.plt.show()


@plotting.customize
def create_event_returns_tear_sheet(factor_data,
                                    returns,
                                    avgretplot=(5, 15),
                                    long_short=True,
                                    group_neutral=False,
                                    std_bar=True,
                                    by_group=False):
    """
    Creates a tear sheet to view the average cumulative returns for a
    factor within a window (pre and post event).

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex Series indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, the factor
        quantile/bin that factor value belongs to and (optionally) the group
        the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    returns : pd.DataFrame
        A DataFrame indexed by date with assets in the columns containing daily
        returns.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    avgretplot: tuple (int, int) - (before, after)
        If not None, plot quantile average cumulative returns
    long_short : bool
        Should this computation happen on a long short portfolio? if so then
        factor returns will be demeaned across the factor universe
    group_neutral : bool
        Should this computation happen on a group neutral portfolio? if so,
        returns demeaning will occur on the group level.
    std_bar : boolean, optional
        Show plots with standard deviation bars, one for each quantile
    by_group : bool
        If True, display graphs separately for each group.
    """

    before, after = avgretplot

    avg_cumulative_returns = perf.average_cumulative_return_by_quantile(
        factor_data,
        returns,
        periods_before=before,
        periods_after=after,
        demeaned=long_short,
        group_adjust=group_neutral,
    )

    num_quantiles = int(factor_data["factor_quantile"].max())

    vertical_sections = 1
    if std_bar:
        vertical_sections += ((num_quantiles - 1) // 2) + 1
    cols = 2 if num_quantiles != 1 else 1
    plotting.plot_quantile_average_cumulative_return(
        avg_cumulative_returns,
        by_quantile=False,
        std_bar=False
    )
    if std_bar:
        plotting.plot_quantile_average_cumulative_return(
            avg_cumulative_returns,
            by_quantile=True,
            std_bar=True
        )

    plt.show()

    if by_group:
        groups = factor_data["group"].unique()
        num_groups = len(groups)
        vertical_sections = ((num_groups - 1) // 2) + 1

        avg_cumret_by_group = perf.average_cumulative_return_by_quantile(
            factor_data,
            returns,
            periods_before=before,
            periods_after=after,
            demeaned=long_short,
            group_adjust=group_neutral,
            by_group=True,
        )

        for group, avg_cumret in avg_cumret_by_group.groupby(level="group"):
            avg_cumret.index = avg_cumret.index.droplevel("group")
            plotting.plot_quantile_average_cumulative_return(
                avg_cumret,
                by_quantile=False,
                std_bar=False,
                title=group
            )

        plt.show()


@plotting.customize
def create_event_study_tear_sheet(factor_data,
                                  returns,
                                  avgretplot=(5, 15),
                                  rate_of_ret=True,
                                  n_bars=50):
    """
    Creates an event study tear sheet for analysis of a specific event.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single event, forward returns for each
        period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
    returns : pd.DataFrame, required only if 'avgretplot' is provided
        A DataFrame indexed by date with assets in the columns containing daily
        returns.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    avgretplot: tuple (int, int) - (before, after), optional
        If not None, plot event style average cumulative returns within a
        window (pre and post event).
    rate_of_ret : bool, optional
        Display rate of return instead of simple return in 'Mean Period Wise
        Return By Factor Quantile' and 'Period Wise Return By Factor Quantile'
        plots
    n_bars : int, optional
        Number of bars in event distribution plot
    """

    long_short = False

    plotting.plot_quantile_statistics_table(factor_data)

    plotting.plot_events_distribution(
        events=factor_data["factor"], num_bars=n_bars
    )
    plt.show()

    if returns is not None and avgretplot is not None:
        create_event_returns_tear_sheet(
            factor_data=factor_data,
            returns=returns,
            avgretplot=avgretplot,
            long_short=long_short,
            group_neutral=False,
            std_bar=True,
            by_group=False,
        )

    factor_returns = perf.factor_returns(
        factor_data, demeaned=False, equal_weight=True
    )

    mean_quant_ret, std_quantile = perf.mean_return_by_quantile(
        factor_data, by_group=False, demeaned=long_short
    )
    if rate_of_ret:
        mean_quant_ret = mean_quant_ret.apply(
            utils.rate_of_return, axis=0, base_period=mean_quant_ret.columns[0]
        )

    mean_quant_ret_bydate, std_quant_daily = perf.mean_return_by_quantile(
        factor_data, by_date=True, by_group=False, demeaned=long_short
    )
    if rate_of_ret:
        mean_quant_ret_bydate = mean_quant_ret_bydate.apply(
            utils.rate_of_return,
            axis=0,
            base_period=mean_quant_ret_bydate.columns[0],
        )

    fr_cols = len(factor_returns.columns)
    vertical_sections = 2 + fr_cols * 1

    plotting.plot_quantile_returns_bar(
        mean_quant_ret, by_group=False, ylim_percentiles=None
    )

    plotting.plot_quantile_returns_violin(
        mean_quant_ret_bydate, ylim_percentiles=(1, 99)
    )

    trading_calendar = factor_data.index.levels[0].freq
    if trading_calendar is None:
        trading_calendar = pd.tseries.offsets.BDay()
        warnings.warn(
            "'freq' not set in factor_data index: assuming business day",
            UserWarning,
        )

    plt.show()

################################################################################


def sd_plot_factor_statistics_table(factor_data):
    plotting.sd_plot_quantile_statistics_table(factor_data)


def sd_get_return_results(factor_data, turnover_Data, ana_mode, plot):
    if ana_mode == 'longonly' or ana_mode == 'benchmark':
        longshort = False
    elif ana_mode == 'longshort':
        longshort = True

    input_periods = utils.get_forward_returns_columns(
        factor_data.columns, require_exact_day_multiple=True,).to_numpy()
    input_periods = np.array(
        utils.timedelta_strings_to_integers(input_periods))

    top_quant = factor_data["factor_quantile"].max()
    bottom_quant = factor_data["factor_quantile"].min()

    if turnover_Data is None:
        turnover_Data = sd_get_turnover_result(factor_data, False)
    quantile_turnover = turnover_Data['quantile_turnover']

    # 分位数分组每日收益
    mean_quant_ret_bydate, std_quant_daily = perf.mean_return_by_quantile(
        factor_data,
        by_date=True,
        by_group=False,
        demeaned=longshort,
        group_adjust=False,
    )
    # 统一为日收益
    mean_quant_rateret_bydate = mean_quant_ret_bydate.apply(
        utils.rate_of_return,
        axis=0,
        base_period=mean_quant_ret_bydate.columns[0],
    )

    # 换手手续费
    expenses = quantile_turnover * -0.0013
    expenses_bydate = expenses.swaplevel(0, 1).reindex(
        mean_quant_rateret_bydate.index).fillna(0)
    mean_quant_rateret_bydate_ae = mean_quant_rateret_bydate + expenses_bydate
    # mean_quant_ret_bydate_ae = mean_quant_ret_bydate + expenses_bydate

    # # 统一为日标准差
    # compstd_quant_daily = std_quant_daily.apply(
    #     utils.std_conversion, axis=0, base_period=std_quant_daily.columns[0]
    # )

    # 计算首尾分组收益差(即多空收益)
    mean_ret_spread_quant, std_spread_quant = perf.compute_mean_returns_spread(
        mean_quant_rateret_bydate,
        top_quant,
        bottom_quant,
        # std_err=compstd_quant_daily,
    )

    mean_ret_spread_quant_ae = expenses.xs(top_quant, level='factor_quantile') + \
        expenses.xs(bottom_quant, level='factor_quantile')

    mean_ret_spread_quant_ae = mean_ret_spread_quant + \
        mean_ret_spread_quant_ae.reindex(mean_ret_spread_quant.index).fillna(0)

    # 因子看板用
    ret_wide = mean_quant_rateret_bydate["1D"].unstack('factor_quantile')
    ret_wide_ae = mean_quant_rateret_bydate_ae["1D"].unstack('factor_quantile')
    cum_ret = ret_wide.apply(perf.cumulative_returns)
    cum_ret_ae = ret_wide_ae.apply(perf.cumulative_returns)

    # 计算年化收益
    days = len(ret_wide)

    if ana_mode == 'longonly':
        mean_quant_rateret = (mean_quant_rateret_bydate.add(1).groupby(mean_quant_rateret_bydate.index.get_level_values('factor_quantile'))
                              .apply('prod')**(243/days)).add(-1)
        mean_quant_rateret_ae = (mean_quant_rateret_bydate_ae.add(1).groupby(mean_quant_rateret_bydate.index.get_level_values('factor_quantile'))
                                 .apply('prod')**(243/days)).add(-1)
    else:
        mean_quant_rateret = (mean_quant_rateret_bydate.groupby(mean_quant_rateret_bydate.index.get_level_values('factor_quantile'))
                              .apply('mean')*(243))
        mean_quant_rateret_ae = (mean_quant_rateret_bydate_ae.groupby(mean_quant_rateret_bydate_ae.index.get_level_values('factor_quantile'))
                                 .apply('mean')*(243))

    # 分组收益表格
    returns_table = pd.DataFrame(columns=mean_quant_ret_bydate.columns)
    returns_table.loc["Annualized Return Top Quantile (%)"] = mean_quant_rateret.loc[top_quant]*100
    returns_table.loc["Annualized Return Bottom Quantile (%)"] = mean_quant_rateret.loc[bottom_quant]*100
    returns_table.loc["Annualized ReturnAE Top Quantile (%)"] = mean_quant_rateret_ae.loc[top_quant]*100
    returns_table.loc["Annualized ReturnAE Bottom Quantile (%)"] = mean_quant_rateret_ae.loc[bottom_quant]*100

    if ana_mode == 'longshort':
        returns_table.loc["Annualized Spread ReturnAE(%)"] = \
            mean_ret_spread_quant_ae.mean() * 100 * 243
        returns_table.loc["Spread ReturnAE max_drawdown(%)"] = \
            ep.max_drawdown(mean_ret_spread_quant_ae).values*100
        returns_table.loc["Spread ReturnAE SharpeRatio"] = \
            ep.sharpe_ratio(mean_ret_spread_quant_ae)/(input_periods**0.5)
    else:
        returns_table.loc["ReturnAE Top Quantile max_drawdown(%)"] = ep.max_drawdown(mean_quant_rateret_bydate_ae.xs(
            top_quant, level='factor_quantile')).values*100
        returns_table.loc["ReturnAE Top Quantile SharpeRatio"] = ep.sharpe_ratio(mean_quant_rateret_bydate_ae.xs(
            top_quant, level='factor_quantile'))/(input_periods**0.5)

    # 分年处理
    returns_table_byYear = pd.DataFrame()
    # 年化收益
    if ana_mode == 'longonly':
        mean_quant_rateret_byYear = (ret_wide.add(1).groupby(ret_wide.index.get_level_values('date').to_period('A')).apply('prod')**(
            243/ret_wide.add(1).groupby(ret_wide.index.get_level_values('date').to_period('A')).apply('count'))).add(-1)
        mean_quant_rateret_byYear_ae = (ret_wide_ae.add(1).groupby(ret_wide_ae.index.get_level_values('date').to_period('A')).apply('prod')**(
            243/ret_wide_ae.add(1).groupby(ret_wide_ae.index.get_level_values('date').to_period('A')).apply('count'))).add(-1)
    else:
        mean_quant_rateret_byYear = ret_wide.groupby(ret_wide.index.get_level_values('date').to_period('A')).apply('mean')*243
        mean_quant_rateret_byYear_ae = ret_wide_ae.groupby(ret_wide_ae.index.get_level_values('date').to_period('A')).apply('mean')*243

    mean_ret_spread_quan_byYeart = mean_ret_spread_quant['1D'].groupby(mean_ret_spread_quant['1D'].index.get_level_values('date').to_period('A')).apply('mean')*243
    mean_ret_spread_quan_byYeart_ae = mean_ret_spread_quant_ae['1D'].groupby(mean_ret_spread_quant_ae['1D'].index.get_level_values('date').to_period('A')).apply('mean')*243

    returns_table_byYear["Annualized Return Top Quantile (%)"] = mean_quant_rateret_byYear[top_quant] * 100
    returns_table_byYear["Annualized ReturnAE Top Quantile (%)"] = mean_quant_rateret_byYear_ae[top_quant] * 100
    returns_table_byYear["Annualized Return Bottom Quantile (%)"] = mean_quant_rateret_byYear[bottom_quant] * 100
    returns_table_byYear["Annualized ReturnAE Bottom Quantile (%)"] = mean_quant_rateret_byYear_ae[bottom_quant] * 100
    returns_table_byYear["Annualized Spread Return(%)"] = mean_ret_spread_quan_byYeart * 100
    returns_table_byYear["Annualized Spread ReturnAE(%)"] = mean_ret_spread_quan_byYeart_ae * 100

    if ana_mode == 'longshort':
        returns_table_byYear["Spread ReturnAE SharpeRatio"] = mean_ret_spread_quant_ae['1D'].groupby(mean_ret_spread_quant_ae['1D'].index.get_level_values('date').to_period('A')).apply(ep.sharpe_ratio)
        returns_table_byYear["Spread ReturnAE max_drawdown"] = mean_ret_spread_quant_ae['1D'].groupby(mean_ret_spread_quant_ae['1D'].index.get_level_values('date').to_period('A')).apply(ep.max_drawdown) * 100

    top_quant_data = mean_quant_rateret_bydate_ae['1D'].xs(top_quant, level='factor_quantile')
    returns_table_byYear["ReturnAE Top Quantile SharpeRatio"] = \
        top_quant_data.groupby(top_quant_data.index.get_level_values('date').to_period('A')).apply(ep.sharpe_ratio)
    returns_table_byYear["ReturnAE Top Quantile max_drawdown(%)"] = \
        top_quant_data.groupby(top_quant_data.index.get_level_values('date').to_period('A')).apply(ep.max_drawdown) * 100

    if plot:
        # 输出分组收益表格

        title = (ana_mode + " AnalysisMode " + "Returns Analysis ")
        print(title)
        utils.print_table(returns_table.apply(lambda x: x.round(2)))
        print("\n")

        # 画分组收益图
        plotting.sd_plot_quantile_returns_bar(
            mean_quant_rateret,
            by_group=False,
            ylim_percentiles=None
        )
        # 画分组累计收益图
        if ana_mode == 'benchmark':
            if "1D" in mean_quant_rateret_bydate:
                plotting.sd_plot_cumulative_excessreturn_by_quantile(
                    mean_quant_rateret_bydate["1D"], period="1D"
                )
                plotting.sd_plot_cumulative_excessreturn(
                    pd.concat([mean_quant_rateret_bydate["1D"].xs(
                        top_quant, level='factor_quantile'),
                        mean_quant_rateret_bydate_ae["1D"].xs(
                            top_quant, level='factor_quantile')], axis=1), period="1D"
                )
        else:
            if "1D" in mean_quant_rateret_bydate:
                plotting.sd_plot_cumulative_returns_by_quantile(
                    mean_quant_rateret_bydate["1D"], period="1D", longshort=longshort
                )
            if "1D" in mean_ret_spread_quant:
                plotting.sd_plot_cumulative_returns_spread(
                    pd.concat([mean_ret_spread_quant["1D"], mean_ret_spread_quant_ae["1D"]], axis=1), period="1D"
                )
        # 输出分年收益表格
        title = (ana_mode + " AnalysisMode " + "Returns Analysis By Year ")
        print(title)
        utils.print_table(returns_table_byYear.apply(lambda x: x.round(2)))
        print("\n")

    ret_Data = {}
    ret_Data['returns_table'] = returns_table
    ret_Data['ret_quant_daily'] = mean_quant_rateret_bydate
    ret_Data['ret_spread_quant_daily'] = mean_ret_spread_quant
    ret_Data['ret_quant_daily_ae'] = mean_quant_rateret_bydate_ae
    ret_Data['ret_spread_quant_daily_ae'] = mean_ret_spread_quant_ae
    ret_Data['cum_ret'] = cum_ret
    ret_Data['ret_wide'] = ret_wide
    ret_Data['cum_ret_ae'] = cum_ret_ae
    ret_Data['ret_wide_ae'] = ret_wide_ae
    ret_Data['returns_table_byYear'] = returns_table_byYear

    ret_Data['quantile_return'] = mean_quant_rateret
    ret_Data['quantile_return_ae'] = mean_quant_rateret_ae

    return ret_Data


def sd_get_information_result(factor_data, plot):
    ic_data = perf.factor_information_coefficient(factor_data, False)
    ic_data_sd_20D = sd_factor_information_coefficient(factor_data, 20)
    ic_data_sd_60D = sd_factor_information_coefficient(factor_data, 60)

    ic_summary_table = pd.DataFrame()
    ic_summary_table["IC Mean"] = ic_data.mean()
    #ic_summary_table["IC Std."] = ic_data.std()
    ic_summary_table["IR"] = \
        ic_data.mean() / ic_data.std()
    ic_summary_table["Good IC Ratio"] = len(ic_data[abs(ic_data['1D']) > 0.02])/len(ic_data)

    ic_summary_table["IC Mean SD_20D"] = ic_data_sd_20D.mean()
    #ic_summary_table["IC Std SD_20D"] = ic_data_sd_20D.std()
    ic_summary_table["IR SD_20D"] = \
        ic_data_sd_20D.mean() / ic_data_sd_20D.std()

    ic_summary_table["IC Mean SD_60D"] = ic_data_sd_60D.mean()
   # ic_summary_table["IC Std SD_60D"] = ic_data_sd_60D.std()
    ic_summary_table["IR SD_60D"] = \
        ic_data_sd_60D.mean() / ic_data_sd_60D.std()
    if plot:
        print("Information Analysis")
        utils.print_table(ic_summary_table.apply(lambda x: x.round(3)).T)
        print("\n")
        # plotting.plot_ic_ts(ic_data)
    return ic_summary_table


def sd_factor_information_coefficient(factor_data, window):
    def src_ic(group):
        f = group['factor_quantile']
        _ic = group[utils.get_forward_returns_columns(factor_data.columns)] \
            .apply(lambda x: stats.spearmanr(x, f)[0])
        return _ic

    def sum_rtn(group):
        group[utils.get_forward_returns_columns(factor_data.columns)] = group[utils.get_forward_returns_columns(
            factor_data.columns)].rolling(window).sum()
        return group

    factor_data = factor_data.copy()
    mean_quant_ret = factor_data.groupby(['date', 'factor_quantile'])[
        utils.get_forward_returns_columns(factor_data.columns)].mean().reset_index()
    mean_quant_ret_window = mean_quant_ret.groupby(
        ['factor_quantile']).apply(sum_rtn).reset_index()

    ic = mean_quant_ret_window.groupby('date').apply(src_ic)

    return ic


def sd_get_turnover_result(factor_data, plot):
    input_periods = utils.get_forward_returns_columns(
        factor_data.columns, require_exact_day_multiple=True,
    ).to_numpy()
    turnover_periods = utils.timedelta_strings_to_integers(input_periods)

    quantile_factor = factor_data["factor_quantile"]

    quantile_turnover = {
        p: pd.concat(
            [
                perf.quantile_turnover(quantile_factor, q, p)
                for q in quantile_factor.sort_values().unique().tolist()
            ],
            axis=1,
        )
        for p in turnover_periods
    }

    autocorrelation = pd.concat(
        [
            perf.factor_rank_autocorrelation(factor_data, period)
            for period in turnover_periods
        ],
        axis=1,
    )

    turnover_table = pd.DataFrame()
    for period in sorted(quantile_turnover.keys()):
        for quantile, p_data in quantile_turnover[period].iteritems():
            turnover_table.loc["Quantile {} Mean Turnover".format(quantile),
                               "{}D".format(period)] = p_data.mean()/period
    auto_corr = pd.DataFrame()
    for period, p_data in autocorrelation.iteritems():
        auto_corr.loc["Mean Factor Rank Autocorrelation",
                      "{}D".format(period)] = p_data.mean()
    if plot:
        print("Turnover Analysis")
        utils.print_table(turnover_table.apply(lambda x: x.round(3)))
        print("\n")
        utils.print_table(auto_corr.apply(lambda x: x.round(3)))
        print("\n")

        for period in turnover_periods:
            if quantile_turnover[period].isnull().all().all():
                continue
            plotting.plot_top_bottom_quantile_turnover(
                quantile_turnover[period], period=period
            )

        for period in autocorrelation:
            if autocorrelation[period].isnull().all():
                continue
            plotting.plot_factor_rank_auto_correlation(
                autocorrelation[period], period=period
            )
    quantile_turnover_temp = pd.DataFrame()
    for t in range(len(turnover_periods)):
        quantile_turnover_temp[input_periods[t]] = quantile_turnover[turnover_periods[t]].stack(
        )/turnover_periods[t]
    quantile_turnover_temp.index.set_names(
        ['date', 'factor_quantile'], inplace=True)

    turnover_Data = {}
    turnover_Data['turnover_table'] = turnover_table

    turnover_Data['quantile_turnover'] = quantile_turnover_temp.fillna(0)
    return turnover_Data


@plotting.customize
def sd_get_Analysis_results(factor_data, ana_mode='longshort', plot=False):
    """
    自定义输出分析结果
    参数：
    ana_mode：longshort测试多空模式, longonly测试long-only模式,benchmark测试对标超额收益模式
    plot：True输出图像

    输出:Dict results,包括关键结果、收益率相关表、换手信息表格、IC值信息表格
    相关子表见后
    results['keyResuls_table']
    results['ret_Data'] = ret_Data
    results['turnover_Data'] = turnover_Data
    results['IC_Data'] = IC_Data

    可用函数包含:

    因子值统计
    tears.sd_plot_factor_statistics_table(factor_data)

    分组收益计算
    tears.sd_get_return_results(factor_data, turnover_Data, ana_mode, plot):
    参数：longshort=True多空模式,False做多模式
    参数：plot=True画图，False不画
    可选参数：turnover_Data换手信息，若不填或为none自动计算
    输出：
    ret_Data['returns_table']收益率结果
    ret_Data['returns_table_byYear']收益率结果按年统计
    ret_Data['ret_quant_daily']分组日收益
    ret_Data['ret_spread_quant_daily']多空收益
    ret_Data['ret_quant_daily_AE']费后分组日收益
    ret_Data['ret_spread_quant_daily_AE']费后多空收益
    ret_Data['cum_ret'] 累计分组收益
    ret_Data['ret_wide'] 分组日收益
    ret_Data['cum_ret_ae'] 累计费后分组收益
    ret_Data['ret_wide_ae'] 费后分组日收益
    ret_Data['quantile_return'] 分组年化收益
    ret_Data['quantile_return_ae'] 费后分组年化收益


    IC值计算
    tears.sd_get_information_result(factor_data,plot=True)
    参数：plot=True画图，False不画
    输出：IC值信息表格 IC_table

    换手计算
    tears.sd_get_turnover_result(factor_data,plot=True)
    参数：plot=True画图，False不画
    输出：换手信息表格，分组日换手
    turnover_Data['turnover_table']
    turnover_Data['quantile_turnover']


    """
    if ana_mode not in ['longonly', 'longshort', 'benchmark']:
        raise ValueError("ana_mode指定错误")

    # 因子值统计板块
    if plot:
        sd_plot_factor_statistics_table(factor_data)

    # 换手分析
    turnover_Data = sd_get_turnover_result(factor_data, plot=plot)

    # 收益分析板块
    # longshort为True时longshort模式分组统计，False时longonly模式
    ret_Data = sd_get_return_results(
        factor_data, turnover_Data=turnover_Data, ana_mode=ana_mode, plot=plot)

    # IC值
    IC_Data = sd_get_information_result(factor_data, plot=plot)

    keyResults_table = pd.DataFrame(columns=ret_Data['returns_table'].columns)
    if ana_mode == 'longshort':
        keyResults_table.loc['Annualized Spread ReturnAE(%)'] = ret_Data['returns_table'].loc[
            "Annualized Spread ReturnAE(%)"]
        keyResults_table.loc['Spread ReturnAE max_drawdown(%)'] = ret_Data[
            'returns_table'].loc["Spread ReturnAE max_drawdown(%)"]
        keyResults_table.loc['Spread ReturnAE SharpeRatio'] = ret_Data['returns_table'].loc["Spread ReturnAE SharpeRatio"]
    elif ana_mode == 'longonly' or ana_mode == 'benchmark':
        keyResults_table.loc['Annualized ReturnAE Top Quantile (%)'] = ret_Data['returns_table'].loc[
            "Annualized ReturnAE Top Quantile (%)"]
        keyResults_table.loc['ReturnAE Top Quantile max_drawdown(%)'] = ret_Data['returns_table'].loc[
            "ReturnAE Top Quantile max_drawdown(%)"]
        keyResults_table.loc['ReturnAE Top Quantile SharpeRatio'] = ret_Data['returns_table'].loc[
            "ReturnAE Top Quantile SharpeRatio"]

    keyResults_table.loc['IC_20'] = IC_Data['IC Mean SD_20D'].T
    keyResults_table.loc['IR_20'] = IC_Data['IR SD_20D'].T
    keyResults_table.loc['IC_60'] = IC_Data['IC Mean SD_60D'].T
    keyResults_table.loc['IR_60'] = IC_Data['IR SD_60D'].T

    keyResults_table.loc['Annualized ReturnAE Top Quantile (%)'] = ret_Data[
        'returns_table'].loc["Annualized ReturnAE Top Quantile (%)"]
    keyResults_table.loc['Annualized ReturnAE Bottom Quantile (%)'] = ret_Data[
        'returns_table'].loc["Annualized ReturnAE Bottom Quantile (%)"]
    keyResults_table.loc['Mean turnover Top Quantile'] = turnover_Data['turnover_table'].iloc[0]
    keyResults_table.loc['Mean turnover Bottom Quantile'] = turnover_Data['turnover_table'].iloc[-1]
    # turnover_Data['turnover_table'].iloc[0]

    results = {}
    results['keyResults'] = keyResults_table
    results['ret_Data'] = ret_Data
    results['turnover_Data'] = turnover_Data
    results['IC_Data'] = IC_Data

    return results
